package ixpce;

import java.io.IOException;

import com.nttdocomo.device.StorageDevice;
import com.nttdocomo.fs.AccessToken;
import com.nttdocomo.fs.DoJaStorageService;
import com.nttdocomo.fs.File;
import com.nttdocomo.fs.Folder;
import com.nttdocomo.io.FileDataOutput;
import com.nttdocomo.io.FileEntity;

class M6502 {
	public int PC;

	public int A, P, X, Y, S;

	public int VF;

	public int IPeriod, ICount;

	public int IRequest;

	public int AfterCLI;

	public int ZF;

	public int IBackup;

	public int NF;

	public int User;

	public int TrapBadOps;

	public int Trap;

	public int Trace;

	public int[] MPR = new int[8];

	private static final int[] Cycles = new int[] {
		8, 7, 3,  4, 6, 4, 6, 7, 3, 2, 2, 0, 7, 5, 7, 6, 
		2, 7, 7,  4, 6, 4, 6, 7, 2, 5, 2, 0, 7, 5, 7, 6, 
		7, 7, 3,  4, 4, 4, 6, 7, 3, 2, 2, 0, 5, 5, 7, 6, 
		2, 7, 7,  0, 4, 4, 6, 7, 2, 5, 2, 0, 5, 5, 7, 6, 
		7, 7, 3,  4, 8, 4, 6, 7, 3, 2, 2, 0, 4, 5, 7, 6, 
		2, 7, 7,  5, 0, 4, 6, 7, 2, 5, 3, 0, 0, 5, 7, 6, 
		7, 7, 2,  0, 4, 4, 6, 7, 3, 2, 2, 0, 7, 5, 7, 6, 
		2, 7, 7, 17, 4, 4, 6, 7, 2, 5, 3, 0, 7, 5, 7, 6, 
		4, 7, 2,  7, 4, 4, 4, 7, 2, 2, 2, 0, 5, 5, 5, 6, 
		2, 7, 7,  8, 4, 4, 4, 7, 2, 5, 2, 0, 5, 5, 5, 6, 
		2, 7, 2,  7, 4, 4, 4, 7, 2, 2, 2, 0, 5, 5, 5, 6, 
		2, 7, 7,  8, 4, 4, 4, 7, 2, 5, 2, 0, 5, 5, 5, 6, 
		2, 7, 2, 17, 4, 4, 6, 7, 2, 2, 2, 0, 5, 5, 7, 6, 
		2, 7, 7, 17, 0, 4, 6, 7, 2, 5, 3, 0, 0, 5, 7, 6, 
		2, 7, 0, 17, 4, 4, 6, 7, 2, 2, 2, 0, 5, 5, 7, 6, 
		2, 7, 7, 17, 2, 4, 6, 7, 2, 5, 3, 0, 0, 5, 7, 0
	};

	private int CycleCountOld = 0;

	private Debug debug = new Debug();

	public void Reset6502(Pce pce) {
		int wk;
		byte page[][] = pce.Page;

		MPR[7] = 0x00;
		pce.bank_set(7, 0x00);
		MPR[6] = 0x05;
		pce.bank_set(6, 0x05);
		MPR[5] = 0x04;
		pce.bank_set(5, 0x04);
		MPR[4] = 0x03;
		pce.bank_set(4, 0x03);
		MPR[3] = 0x02;
		pce.bank_set(3, 0x02);
		MPR[2] = 0x01;
		pce.bank_set(2, 0x01);
		MPR[1] = 0xF8;
		pce.bank_set(1, 0xF8);
		MPR[0] = 0xFF;
		pce.bank_set(0, 0xFF);

		A = X = Y = 0x00;
		P = 0x04;
		NF = VF = 0;
		ZF = 0xFF;
		S = 0xFF;
		PC = (PC & 0xFF00) | (page[(wk = (0xFFFE)) >> 13][wk - (wk & ~0x1FFF)] & 0xFF);
		PC = (PC & 0xFF) | (((page[(wk = (0xFFFE + 1)) >> 13][wk - (wk & ~0x1FFF)] & 0xFF)) << 8);
		ICount = IPeriod;
		IRequest = 0;
		AfterCLI = 0;
		User = 0;
	}

	public void Int6502(Pce pce, int Type) {
		int wk;
		int J = 0;
		byte page[][] = pce.Page;
		byte ram[] = pce.RAM;

		if ((Type == 2) || ((P & 0x04) == 0)) {
			ICount -= 7;
			ram[0x100 + S] = (byte)((PC) >> 8);
			S = (S - 1) & 0xFF;
			ram[0x100 + S] = (byte)((PC) & 0xFF);
			S = (S - 1) & 0xFF;
			ram[0x100 + S] = (byte)(((P & ~(0x10 | 0x20)) & ~(0x80 | 0x40 | 0x02)) | (NF & 0x80) | (VF & 0x40) | ((ZF != 0) ? 0 : 0x02));
			S = (S - 1) & 0xFF;
			P &= ~0x08;
			if (Type == 2) {
				J = 0xFFFC;
			} else {
				P |= 0x04;
				switch (Type) {
				case 1:
					J = 0xFFF8;
					break;
				case 8:
					J = 0xFFF6;
					break;
				case 4:
					J = 0xFFFA;
					break;
				}
			}
			PC = (PC & 0xFF00) | (page[(wk = (J)) >> 13][wk - (wk & ~0x1FFF)] & 0xFF);
			PC = (PC & 0xFF) | (((page[(wk = (J + 1)) >> 13][wk - (wk & ~0x1FFF)] & 0xFF)) << 8);
		} else {
			IRequest |= Type;
		}
	}
	
	private FileEntity fe = null;
	private FileDataOutput out = null;
	
	public void InitSDLog() {
		StorageDevice sd = StorageDevice.getInstance("/ext0");
		Folder folder;
		try {
			folder = sd.getFolder((AccessToken)DoJaStorageService.getAccessToken(0,DoJaStorageService.SHARE_APPLICATION ));
			try
			{
				folder.getFile("doja_log.txt").delete();
			} catch (IOException ex) {
			}
			File file = folder.createFile("doja_log.txt");
			fe = file.open(File.MODE_WRITE_ONLY);			 
			out = fe.openDataOutput();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
	
	int count = 0;
	
	public void CloseLog() {
		try {
			out.close();
			fe.close();		
		} catch (IOException e) {
			e.printStackTrace();
		}				
	}
	
	public int Run6502(Pce pce) {
		int J = 0;
		int K = 0;
		int wk;
		int I;
		int pc = PC;
		int a = A;
		int p = P;
		int x = X;
		int y = Y;
		int s = S;
		final byte page[][] = pce.Page;
		final byte ram[] = pce.RAM;
		
		int cycle;
		final int msk = ~0x1FFF;
		
		for (;;) {
			//try {
			
			I = (page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF);
			
			//Debug.addDebugMessage(I, pc, a, p, x, y, s, User);

			cycle = Cycles[I];
			switch (I) {

			case 0xD4:
				break;
				
			case 0x54:
				break;

			case 0x44:
				ram[0x100 + s] = (byte)(pc >> 8);
				s = (s - 1) & 0xFF;
				ram[0x100 + s] = (byte)(pc & 0xFF);
				s = (s - 1) & 0xFF;
			case 0x80:
				pc += (byte) (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF) + 1;
				break;

			case 0x7C:
				K = (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF);
				K |= (((page[(wk = (pc + 1)) >> 13][wk - (wk & msk)] & 0xFF)) << 8);
				pc += 2;
				K += x;
				pc = (page[(wk = K) >> 13][wk - (wk & msk)] & 0xFF);
				pc |= (((page[(wk = (K + 1)) >> 13][wk - (wk & msk)] & 0xFF)) << 8);
				break;

			case 0xDA:
				ram[0x100 + s] = (byte)x;
				s = (s - 1) & 0xFF;
				break;
				
			case 0x5A:
				ram[0x100 + s] = (byte)y;
				s = (s - 1) & 0xFF;
				break;
				
			case 0xFA:
				s = (s + 1) & 0xFF;
				x = (ram[0x100 + s] & 0xFF);
				ZF = NF = x;
				break;
				
			case 0x7A:
				s = (s + 1) & 0xFF;
				y = (ram[0x100 + s] & 0xFF);
				ZF = NF = y;
				break;

			case 0x62:
				a = 0;
				break;
				
			case 0x82:
				x = 0;
				break;
				
			case 0xC2:
				y = 0;
				break;

			case 0x02:
				I = x;
				x = y;
				y = I;
				break;
				
			case 0x22:
				I = a;
				a = x;
				x = I;
				break;
				
			case 0x42:
				I = a;
				a = y;
				y = I;
				break;

			case 0x3A:
				a = (a - 1) & 0xFF;
				ZF = NF = a;
				break;
				
			case 0x1A:
				a = (a + 1) & 0xFF;
				ZF = NF = a;
				break;

			case 0x72:
				K = ((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF));
				J = (ram[K] & 0xFF);
				J |= (((ram[K + 1] & 0xFF)) << 8);
				I = pce._Rd6502(J);
				if ((p & 0x08) != 0) {
					K = (a & 0x0F) + (I & 0x0F) + (p & 0x01);
					K |= (((a >> 4) + (I >> 4)) << 8);
					if ((K & 0xFF) > 9) {
						K = (K & 0xFF) + 6;
						K |= (((K >> 8) + 1) << 8);
					}
					if ((K >> 8) > 9)
						K = (K & 0xFF) | (((K >> 8) + 6) << 8);
					a = (K & 0x0F) | ((K >> 8) << 4);
					p = (p & ~0x01) | ((K >> 8) > 15 ? 0x01 : 0);
					ZF = NF = a;
					cycle++;
				} else {
					K = a + I + (p & 0x01);
					p &= ~0x01;
					p |= (((K >> 8) != 0) ? 0x01 : 0);
					VF = (~(a ^ I) & (a ^ (K & 0xFF))) >> 1;
					ZF = NF = (K & 0xFF);
					a = (K & 0xFF);
				}
				break;
				
			case 0x32:
				K = ((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF));
				J = (ram[K] & 0xFF);
				J |= (((ram[K + 1] & 0xFF)) << 8);
				I = pce._Rd6502(J);
				a &= I;
				ZF = NF = a;
				break;
				
			case 0xD2:
				K = ((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF));
				J = (ram[K] & 0xFF);
				J |= (((ram[K + 1] & 0xFF)) << 8);
				I = pce._Rd6502(J);
				K = a - I;
				p &= ~0x01;
				p |= (((K >> 8) != 0) ? 0 : 0x01);
				ZF = NF = (K & 0xFF);
				break;
				
			case 0x52:
				K = ((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF));
				J = (ram[K] & 0xFF);
				J |= (((ram[K + 1] & 0xFF)) << 8);
				I = pce._Rd6502(J);
				a ^= I;
				ZF = NF = a;
				break;
				
			case 0xB2:
				K = ((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF));
				J = (ram[K] & 0xFF);
				J |= (((ram[K + 1] & 0xFF)) << 8);
				a = pce._Rd6502(J);
				ZF = NF = a;
				break;
				
			case 0x12:
				K = ((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF));
				J = (ram[K] & 0xFF);
				J |= (((ram[K + 1] & 0xFF)) << 8);
				I = pce._Rd6502(J);
				a |= I;
				ZF = NF = a;
				break;
				
			case 0x92:
				K = ((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF));
				J = (ram[K] & 0xFF);
				J |= (((ram[K + 1] & 0xFF)) << 8);
				pce._Wr6502(J, a);
				break;
				
			case 0xF2:
				K = ((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF));
				J = (ram[K] & 0xFF);
				J |= (((ram[K + 1] & 0xFF)) << 8);
				I = pce._Rd6502(J);
				if ((p & 0x08) != 0) {
					K = (K & 0xFF00) | (a & 0x0F) - (I & 0x0F) - (~p & 0x01);
					if (((K & 0xFF) & 0x10) != 0)
						K = (K & 0xFF00) | (K & 0xFF) - 6;
					K = (K & 0xFF) | (((a >> 4) - (I >> 4) - ((((K & 0xFF) & 0x10) == 0x10) ? 1 : 0)) << 8);
					if (((K >> 8) & 0x10) != 0)
						K = (K & 0xFF) | (((K >> 8) - 6) << 8);
					a = (K & 0x0F) | ((K >> 8) << 4);
					p = (p & ~0x01) | ((((K >> 8) & 0x10) != 0) ? 0 : 0x01);
					ZF = NF = a;
					cycle++;
				} else {
					K = a - I - (~p & 0x01);
					p &= ~0x01;
					p |= (((K >> 8) != 0) ? 0 : 0x01);
					VF = ((a ^ I) & (a ^ (K & 0xFF))) >> 1;
					ZF = NF = (K & 0xFF);
					a = (K & 0xFF);
				}
				break;

			case 0x89:
				I = (page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF);
				NF = VF = I;
				ZF = I & a;
				break;
				
			case 0x34:
				I = (ram[(byte) ((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF) + x)] & 0xFF);
				NF = VF = I;
				ZF = I & a;
				break;
				
			case 0x3C:
				J = (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF);
				J |=  (((page[(wk = (pc + 1)) >> 13][wk - (wk & msk)] & 0xFF)) << 8);
				pc += 2;
				J += x;
				I = pce._Rd6502(J);
				NF = VF = I;
				ZF = I & a;
				break;

			case 0x64:
				ram[((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF))] = 0x00;
				break;
				
			case 0x74:
				ram[((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF) + x) & 0xFF] = 0x00;
				break;
				
			case 0x9C:
				J = (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF);
				J |= (((page[(wk = (pc + 1)) >> 13][wk - (wk & msk)] & 0xFF)) << 8);
				pc += 2;
				pce._Wr6502(J, 0x00);
				break;
				
			case 0x9E:
				J = (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF);
				J |= (((page[(wk = (pc + 1)) >> 13][wk - (wk & msk)] & 0xFF)) << 8);
				pc += 2;
				J += x;
				pce._Wr6502(J, 0x00);
				break;
				
			case 0xF4:
				I = (page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF);
				cycle += Cycles[I] + 3;
				switch (I) {
				case 0x65:
					I = (ram[((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF))] & 0xFF);
					{
						int index = (x);
						if ((p & 0x08) != 0) {
							K = (ram[index] & 0x0F) + (I & 0x0F) + (p & 0x01);
							K |= ((((ram[index] & 0xFF) >> 4) + (I >> 4)) << 8);
							if ((K & 0xFF) > 9) {
								K = (K & 0xFF) + 6;
								K |= (((K >> 8) + 1) << 8);
							}
							if ((K >> 8) > 9)
								K = (K & 0xFF) | (((K >> 8) + 6) << 8);
							ram[index] = (byte)(((K & 0xFF) & 0x0F) | ((K >> 8) << 4));
							p = (p & ~0x01) | ((K >> 8) > 15 ? 0x01 : 0);
							ZF = NF = (ram[index] & 0xFF);
							cycle++;
						} else {
							K = (ram[index] & 0xFF) + I + (p & 0x01);
							p &= ~0x01;
							p |= (((K >> 8) != 0) ? 0x01 : 0);
							VF = (~((ram[index] & 0xFF) ^ I) & ((ram[index] & 0xFF) ^ (K & 0xFF))) >> 1;
							ZF = NF = (K & 0xFF);
							ram[index] = (byte)(K & 0xFF);
						}
					}
					break;
					
				case 0x6D:
					J = (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF);
					J |= (((page[(wk = (pc + 1)) >> 13][wk - (wk & msk)] & 0xFF)) << 8);
					pc += 2;
					I = pce._Rd6502(J);
					{
						int index = (x);
						if ((p & 0x08) != 0) {
							K = (ram[index] & 0x0F) + (I & 0x0F) + (p & 0x01);
							K |= ((((ram[index] & 0xFF) >> 4) + (I >> 4)) << 8);
							if ((K & 0xFF) > 9) {
								K = (K & 0xFF) + 6;
								K |= (((K >> 8) + 1) << 8);
							}
							if ((K >> 8) > 9)
								K = (K & 0xFF) | (((K >> 8) + 6) << 8);
							ram[index] = (byte)(((K & 0xFF) & 0x0F) | ((K >> 8) << 4));
							p = (p & ~0x01) | ((K >> 8) > 15 ? 0x01 : 0);
							ZF = NF = (ram[index] & 0xFF);
							cycle++;
						} else {
							K = (ram[index] & 0xFF) + I + (p & 0x01);
							p &= ~0x01;
							p |= (((K >> 8) != 0) ? 0x01 : 0);
							VF = (~((ram[index] & 0xFF) ^ I) & ((ram[index] & 0xFF) ^ (K & 0xFF))) >> 1;
							ZF = NF = (K & 0xFF);
							ram[index] = (byte)(K & 0xFF);
						}
					}
					break;
					
				case 0x69:
					I = (page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF);
					{
						int index = (x);
						if ((p & 0x08) != 0) {
							K = (ram[index] & 0x0F) + (I & 0x0F) + (p & 0x01);
							K |= ((((ram[index] & 0xFF) >> 4) + (I >> 4)) << 8);
							if ((K & 0xFF) > 9) {
								K = (K & 0xFF) + 6;
								K |= (((K >> 8) + 1) << 8);
							}
							if ((K >> 8) > 9)
								K = (K & 0xFF) | (((K >> 8) + 6) << 8);
							ram[index] = (byte)(((K & 0xFF) & 0x0F) | ((K >> 8) << 4));
							p = (p & ~0x01) | ((K >> 8) > 15 ? 0x01 : 0);
							ZF = NF = (ram[index] & 0xFF);
							cycle++;
						} else {
							K = (ram[index] & 0xFF) + I + (p & 0x01);
							p &= ~0x01;
							p |= (((K >> 8) != 0) ? 0x01 : 0);
							VF = (~((ram[index] & 0xFF) ^ I) & ((ram[index] & 0xFF) ^ (K & 0xFF))) >> 1;
							ZF = NF = (K & 0xFF);
							ram[index] = (byte)(K & 0xFF);
						}
					}
					break;
					
				case 0x75:
					I = (ram[(byte) ((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF) + x)] & 0xFF);
					{
						int index = (x);
						if ((p & 0x08) != 0) {
							K = (ram[index] & 0x0F) + (I & 0x0F) + (p & 0x01);
							K |= ((((ram[index] & 0xFF) >> 4) + (I >> 4)) << 8);
							if ((K & 0xFF) > 9) {
								K = (K & 0xFF) + 6;
								K |= (((K >> 8) + 1) << 8);
							}
							if ((K >> 8) > 9)
								K = (K & 0xFF) | (((K >> 8) + 6) << 8);
							ram[index] = (byte)(((K & 0xFF) & 0x0F) | ((K >> 8) << 4));
							p = (p & ~0x01) | ((K >> 8) > 15 ? 0x01 : 0);
							ZF = NF = (ram[index] & 0xFF);
							cycle++;
						} else {
							K = (ram[index] & 0xFF) + I + (p & 0x01);
							p &= ~0x01;
							p |= (((K >> 8) != 0) ? 0x01 : 0);
							VF = (~((ram[index] & 0xFF) ^ I) & ((ram[index] & 0xFF) ^ (K & 0xFF))) >> 1;
							ZF = NF = (K & 0xFF);
							ram[index] = (byte)(K & 0xFF);
						}
					}
					break;
					
				case 0x79:
					J = (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF);
					J |= (((page[(wk = (pc + 1)) >> 13][wk - (wk & msk)] & 0xFF)) << 8);
					pc += 2;
					J += y;
					I = pce._Rd6502(J);
					{
						int index = (x);
						if ((p & 0x08) != 0) {
							K = (ram[index] & 0x0F) + (I & 0x0F) + (p & 0x01);
							K |= ((((ram[index] & 0xFF) >> 4) + (I >> 4)) << 8);
							if ((K & 0xFF) > 9) {
								K = (K & 0xFF) + 6;
								K |= (((K >> 8) + 1) << 8);
							}
							if ((K >> 8) > 9)
								K = (K & 0xFF) | (((K >> 8) + 6) << 8);
							ram[index] = (byte)(((K & 0xFF) & 0x0F) | ((K >> 8) << 4));
							p = (p & ~0x01) | ((K >> 8) > 15 ? 0x01 : 0);
							ZF = NF = (ram[index] & 0xFF);
							cycle++;
						} else {
							K = (ram[index] & 0xFF) + I + (p & 0x01);
							p &= ~0x01;
							p |= (((K >> 8) != 0) ? 0x01 : 0);
							VF = (~((ram[index] & 0xFF) ^ I) & ((ram[index] & 0xFF) ^ (K & 0xFF))) >> 1;
							ZF = NF = (K & 0xFF);
							ram[index] = (byte)(K & 0xFF);
						}
					}
					break;
					
				case 0x7D:
					J = (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF);
					J |= (((page[(wk = (pc + 1)) >> 13][wk - (wk & msk)] & 0xFF)) << 8);
					pc += 2;
					J += x;
					I = pce._Rd6502(J);
					{
						int index = (x);
						if ((p & 0x08) != 0) {
							K = (ram[index] & 0x0F) + (I & 0x0F) + (p & 0x01);
							K |= ((((ram[index] & 0xFF) >> 4) + (I >> 4)) << 8);
							if ((K & 0xFF) > 9) {
								K = (K & 0xFF) + 6;
								K |= (((K >> 8) + 1) << 8);
							}
							if ((K >> 8) > 9)
								K = (K & 0xFF) | (((K >> 8) + 6) << 8);
							ram[index] = (byte)(((K & 0xFF) & 0x0F) | ((K >> 8) << 4));
							p = (p & ~0x01) | ((K >> 8) > 15 ? 0x01 : 0);
							ZF = NF = (ram[index] & 0xFF);
							cycle++;
						} else {
							K = (ram[index] & 0xFF) + I + (p & 0x01);
							p &= ~0x01;
							p |= (((K >> 8) != 0) ? 0x01 : 0);
							VF = (~((ram[index] & 0xFF) ^ I) & ((ram[index] & 0xFF) ^ (K & 0xFF))) >> 1;
							ZF = NF = (K & 0xFF);
							ram[index] = (byte)(K & 0xFF);
						}
					}
					break;
					
				case 0x61:
					K = ((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF) + x) & 0xFF;
					J = (ram[K] & 0xFF);
					J |= (((ram[K + 1] & 0xFF)) << 8);
					I = pce._Rd6502(J);
					{
						int index = (x);
						if ((p & 0x08) != 0) {
							K = (ram[index] & 0x0F) + (I & 0x0F)	+ (p & 0x01);
							K |= ((((ram[index] & 0xFF) >> 4) + (I >> 4)) << 8);
							if ((K & 0xFF) > 9) {
								K = (K & 0xFF) + 6;
								K |= (((K >> 8) + 1) << 8);
							}
							if ((K >> 8) > 9)
								K = (K & 0xFF) | (((K >> 8) + 6) << 8);
							ram[index] = (byte)(((K & 0xFF) & 0x0F) | ((K >> 8) << 4));
							p = (p & ~0x01) | ((K >> 8) > 15 ? 0x01 : 0);
							ZF = NF = (ram[index] & 0xFF);
							cycle++;
						} else {
							K = (ram[index] & 0xFF) + I + (p & 0x01);
							p &= ~0x01;
							p |= (((K >> 8) != 0) ? 0x01 : 0);
							VF = (~((ram[index] & 0xFF) ^ I) & ((ram[index] & 0xFF) ^ (K & 0xFF))) >> 1;
							ZF = NF = (K & 0xFF);
							ram[index] = (byte)(K & 0xFF);
						}
					}
					break;
					
				case 0x71:
					K = ((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF));
					J = (ram[K] & 0xFF);
					J |= (((ram[K + 1] & 0xFF)) << 8);
					J += y;
					I = pce._Rd6502(J);
					{
						int index = (x);
						if ((p & 0x08) != 0) {
							K = (ram[index] & 0x0F) + (I & 0x0F)	+ (p & 0x01);
							K |= ((((ram[index] & 0xFF) >> 4) + (I >> 4)) << 8);
							if ((K & 0xFF) > 9) {
								K = (K & 0xFF) + 6;
								K |= (((K >> 8) + 1) << 8);
							}
							if ((K >> 8) > 9)
								K = (K & 0xFF) | (((K >> 8) + 6) << 8);
							ram[index] = (byte)(((K & 0xFF) & 0x0F) | ((K >> 8) << 4));
							p = (p & ~0x01) | ((K >> 8) > 15 ? 0x01 : 0);
							ZF = NF = (ram[index] & 0xFF);
							cycle++;
						} else {
							K = (ram[index] & 0xFF) + I + (p & 0x01);
							p &= ~0x01;
							p |= (((K >> 8) != 0) ? 0x01 : 0);
							VF = (~((ram[index] & 0xFF) ^ I) & ((ram[index] & 0xFF) ^ (K & 0xFF))) >> 1;
							ZF = NF = (K & 0xFF);
							ram[index] = (byte)(K & 0xFF);
						}
					}
					break;
					
				case 0x72:
					K = ((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF));
					J = (ram[K] & 0xFF);
					J |= (((ram[K + 1] & 0xFF)) << 8);
					I = pce._Rd6502(J);
					{
						int index = (x);
						if ((p & 0x08) != 0) {
							K = (ram[index] & 0x0F) + (I & 0x0F)	+ (p & 0x01);
							K |= ((((ram[index] & 0xFF) >> 4) + (I >> 4)) << 8);
							if ((K & 0xFF) > 9) {
								K = (K & 0xFF) + 6;
								K |= (((K >> 8) + 1) << 8);
							}
							if ((K >> 8) > 9)
								K = (K & 0xFF) | (((K >> 8) + 6) << 8);
							ram[index] = (byte)(((K & 0xFF) & 0x0F) | ((K >> 8) << 4));
							p = (p & ~0x01) | ((K >> 8) > 15 ? 0x01 : 0);
							ZF = NF = (ram[index] & 0xFF);
							cycle++;
						} else {
							K = (ram[index] & 0xFF) + I + (p & 0x01);
							p &= ~0x01;
							p |= (((K >> 8) != 0) ? 0x01 : 0);
							VF = (~((ram[index] & 0xFF) ^ I) & ((ram[index] & 0xFF) ^ (K & 0xFF))) >> 1;
							ZF = NF = (K & 0xFF);
							ram[index] = (byte)(K & 0xFF);
						}
					}
					break;

				case 0x25:
					I = (ram[((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF))] & 0xFF);
					ram[x] = (byte)((ram[x] & 0xFF) & I);
					ZF = NF = (ram[x] & 0xFF);
					break;
					
				case 0x2D:
					J = (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF);
					J |= (((page[(wk = (pc + 1)) >> 13][wk - (wk & msk)] & 0xFF)) << 8);
					pc += 2;
					I = pce._Rd6502(J);
					ram[x] = (byte)((ram[x] & 0xFF) & I);
					ZF = NF = (ram[x] & 0xFF);
					break;
					
				case 0x29:
					I = (page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF);
					ram[x] = (byte)((ram[x] & 0xFF) & I);
					ZF = NF = (ram[x] & 0xFF);
					break;
					
				case 0x35:
					I = (ram[((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF) + x) & 0xFF] & 0xFF);
					ram[x] = (byte)((ram[x] & 0xFF) & I);
					ZF = NF = (ram[x] & 0xFF);
					break;
					
				case 0x39:
					J = (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF);
					J |= (((page[(wk = (pc + 1)) >> 13][wk - (wk & msk)] & 0xFF)) << 8);
					pc += 2;
					J += y;
					I = pce._Rd6502(J);
					ram[x] = (byte)((ram[x] & 0xFF) & I);
					ZF = NF = (ram[x] & 0xFF);
					break;
					
				case 0x3D:
					J = (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF);
					J |= (((page[(wk = (pc + 1)) >> 13][wk - (wk & msk)] & 0xFF)) << 8);
					pc += 2;
					J += x;
					I = pce._Rd6502(J);
					ram[x] = (byte)((ram[x] & 0xFF) & I);
					ZF = NF = (ram[x] & 0xFF);
					break;
					
				case 0x21:
					K = ((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF) + x) & 0xFF;
					J = (ram[K] & 0xFF);
					J |= (((ram[K + 1] & 0xFF)) << 8);
					I = pce._Rd6502(J);
					ram[x] = (byte)((ram[x] & 0xFF) & I);
					ZF = NF = (ram[x] & 0xFF);
					break;
					
				case 0x31:
					K = ((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF));
					J = (ram[K] & 0xFF);
					J |= (((ram[K + 1] & 0xFF)) << 8);
					J += y;
					I = pce._Rd6502(J);
					ram[x] = (byte)((ram[x] & 0xFF) & I);
					ZF = NF = (ram[x] & 0xFF);
					break;
					
				case 0x32:
					K = ((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF));
					J = (ram[K] & 0xFF);
					J |= (((ram[K + 1] & 0xFF)) << 8);
					I = pce._Rd6502(J);
					ram[x] = (byte)((ram[x] & 0xFF) & I);
					ZF = NF = (ram[x] & 0xFF);
					break;

				case 0x45:
					I = (ram[((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF))] & 0xFF);
					ram[x] = (byte)((ram[x] & 0xFF) ^ I);
					ZF = NF = (ram[x] & 0xFF);
					break;
					
				case 0x4D:
					J = (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF);
					J |= (((page[(wk = (pc + 1)) >> 13][wk - (wk & msk)] & 0xFF)) << 8);
					pc += 2;
					I = pce._Rd6502(J);
					ram[x] = (byte)((ram[x] & 0xFF) ^ I);
					ZF = NF = (ram[x] & 0xFF);
					break;
					
				case 0x49:
					I = (page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF);
					ram[x] = (byte)((ram[x] & 0xFF) ^ I);
					ZF = NF = (ram[x] & 0xFF);
					break;
					
				case 0x55:
					I = (ram[((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF) + x) & 0xFF] & 0xFF);
					ram[x] = (byte)((ram[x] & 0xFF) ^ I);
					ZF = NF = (ram[x] & 0xFF);
					break;
					
				case 0x59:
					J = (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF);
					J |= (((page[(wk = (pc + 1)) >> 13][wk - (wk & msk)] & 0xFF)) << 8);
					pc += 2;
					J += y;
					I = pce._Rd6502(J);
					ram[x] = (byte)((ram[x] & 0xFF) ^ I);
					ZF = NF = (ram[x] & 0xFF);
					break;
					
				case 0x5D:
					J = (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF);
					J |= (((page[(wk = (pc + 1)) >> 13][wk - (wk & msk)] & 0xFF)) << 8);
					pc += 2;
					J += x;
					I = pce._Rd6502(J);
					ram[x] = (byte)((ram[x] & 0xFF) ^ I);
					ZF = NF = (ram[x] & 0xFF);
					break;
					
				case 0x41:
					K = ((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF) + x) & 0xFF;
					J = (ram[K] & 0xFF);
					J |= (((ram[K + 1] & 0xFF)) << 8);
					I = pce._Rd6502(J);
					ram[x] = (byte)((ram[x] & 0xFF) ^ I);
					ZF = NF = (ram[x] & 0xFF);
					break;
					
				case 0x51:
					K = ((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF));
					J = (ram[K] & 0xFF);
					J |= (((ram[K + 1] & 0xFF)) << 8);
					J += y;
					I = pce._Rd6502(J);
					ram[x] = (byte)((ram[x] & 0xFF) ^ I);
					ZF = NF = (ram[x] & 0xFF);
					break;
					
				case 0x52:
					K = ((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF));
					J = (ram[K] & 0xFF);
					J |= (((ram[K + 1] & 0xFF)) << 8);
					I = pce._Rd6502(J);
					ram[x] = (byte)((ram[x] & 0xFF) ^ I);
					ZF = NF = (ram[x] & 0xFF);
					break;

				case 0x05:
					I = (ram[((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF))] & 0xFF);
					ram[x] = (byte)((ram[x] & 0xFF) | I);
					ZF = NF = (ram[x] & 0xFF);
					break;
					
				case 0x0D:
					J = (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF);
					J |= (((page[(wk = (pc + 1)) >> 13][wk - (wk & msk)] & 0xFF)) << 8);
					pc += 2;
					I = pce._Rd6502(J);
					ram[x] = (byte)((ram[x] & 0xFF) | I);
					ZF = NF = (ram[x] & 0xFF);
					break;
					
				case 0x09:
					I = (page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF);
					ram[x] = (byte)((ram[x] & 0xFF) | I);
					ZF = NF = (ram[x] & 0xFF);
					break;
					
				case 0x15:
					I = (ram[((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF) + x) & 0xFF] & 0xFF);
					ram[x] = (byte)((ram[x] & 0xFF) | I);
					ZF = NF = (ram[x] & 0xFF);
					break;
					
				case 0x19:
					J = (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF);
					J |= (((page[(wk = (pc + 1)) >> 13][wk - (wk & msk)] & 0xFF)) << 8);
					pc += 2;
					J += y;
					I = pce._Rd6502(J);
					ram[x] = (byte)((ram[x] & 0xFF) | I);
					ZF = NF = (ram[x] & 0xFF);
					break;
					
				case 0x1D:
					J = (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF);
					J |= (((page[(wk = (pc + 1)) >> 13][wk - (wk & msk)] & 0xFF)) << 8);
					pc += 2;
					J += x;
					I = pce._Rd6502(J);
					ram[x] = (byte)((ram[x] & 0xFF) | I);
					ZF = NF = (ram[x] & 0xFF);
					break;
					
				case 0x01:
					K = ((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF) + x) & 0xFF;
					J = (ram[K] & 0xFF);
					J |= (((ram[K + 1] & 0xFF)) << 8);
					I = pce._Rd6502(J);
					ram[x] = (byte)((ram[x] & 0xFF) | I);
					ZF = NF = (ram[x] & 0xFF);
					break;
					
				case 0x11:
					K = ((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF));
					J = (ram[K] & 0xFF);
					J |= (((ram[K + 1] & 0xFF)) << 8);
					J += y;
					I = pce._Rd6502(J);
					ram[x] = (byte)((ram[x] & 0xFF) | I);
					ZF = NF = (ram[x] & 0xFF);
					break;
					
				case 0x12:
					K = ((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF));
					J = (ram[K] & 0xFF);
					J |= (((ram[K + 1] & 0xFF)) << 8);
					I = pce._Rd6502(J);
					ram[x] = (byte)((ram[x] & 0xFF) | I);
					ZF = NF = (ram[x] & 0xFF);
					break;

				default:
					debug.TRACE("no sense SET\n");
					cycle -= Cycles[I];
					pc--;
					break;
				}
				break;

			case 0x03:
				pce.IO_write(0, (page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF));
				break;
				
			case 0x13:
				pce.IO_write(2, (page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF));
				break;
				
			case 0x23:
				pce.IO_write(3, (page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF));
				break;

			case 0x43:
				I = (page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF);
				{
					int i;
					for (i = 0; i < 8; i++, I >>= 1) {
						if ((I & 1) != 0)
							break;
					}
					a = MPR[i];
				}
				break;

			case 0x53:
				I = (page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF);
				{
					int i;
					for (i = 0; i < 8; i++, I >>= 1) {
						if ((I & 1) != 0) {
							MPR[i] = a;
							pce.bank_set(i, a);
						}
					}
				}
				break;

			case 0xC3:
				{
					int src, dist, length;
					src = ((page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF) | ((page[(wk + 1) >> 13][wk + 1 - (wk & msk)] & 0xFF) << 8));
					pc += 2;
					dist = ((page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF) | ((page[(wk + 1) >> 13][wk + 1 - (wk & msk)] & 0xFF) << 8));
					pc += 2;
					length = ((page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF) | ((page[(wk + 1) >> 13][wk + 1 - (wk & msk)] & 0xFF) << 8));
					pc += 2;
					cycle = length * 6;
					do {
						pce._Wr6502(dist--, pce._Rd6502(src));
						src--;
					} while (--length > 0);
				}
				break;

			case 0x73:
				{
					int src, dist, length;
					src = ((page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF) | ((page[(wk + 1) >> 13][wk + 1 - (wk & msk)] & 0xFF) << 8));
					pc += 2;
					dist = ((page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF) | ((page[(wk + 1) >> 13][wk + 1 - (wk & msk)] & 0xFF) << 8));
					pc += 2;
					length = ((page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF) | ((page[(wk + 1) >> 13][wk + 1 - (wk & msk)] & 0xFF) << 8));
					pc += 2;
					cycle = length * 6;
					do {
						pce._Wr6502(dist++, pce._Rd6502(src));
						src++;
					} while (--length > 0);
				}
				break;

			case 0xE3:
				{
					int src, dist, length;
					src = ((page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF) | ((page[(wk + 1) >> 13][wk + 1 - (wk & msk)] & 0xFF) << 8));
					pc += 2;
					dist = ((page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF) | ((page[(wk + 1) >> 13][wk + 1 - (wk & msk)] & 0xFF) << 8));
					pc += 2;
					length = ((page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF) | ((page[(wk + 1) >> 13][wk + 1 - (wk & msk)] & 0xFF) << 8));
					pc += 2;
					cycle = length * 6;
					do {
						pce._Wr6502(dist, pce._Rd6502(src));
						src++;
						if (--length == 0)
							break;
						pce._Wr6502(dist + 1, pce._Rd6502(src));
						src++;
					} while (--length > 0);
				}
				break;

			case 0xF3:
				{
					int src, dist, length;
					src = ((page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF) | ((page[(wk + 1) >> 13][wk + 1 - (wk & msk)] & 0xFF) << 8));
					pc += 2;
					dist = ((page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF) | ((page[(wk + 1) >> 13][wk + 1 - (wk & msk)] & 0xFF) << 8));
					pc += 2;
					length = ((page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF) | ((page[(wk + 1) >> 13][wk + 1 - (wk & msk)] & 0xFF) << 8));
					pc += 2;
					cycle = length * 6;
					do {
						pce._Wr6502(dist++, pce._Rd6502(src));
						if (--length == 0)
							break;
						pce._Wr6502(dist++, pce._Rd6502(src + 1));
					} while (--length > 0);
				}
				break;

			case 0xD3:
				{
					int src, dist, length;
					src = ((page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF) | ((page[(wk + 1) >> 13][wk + 1 - (wk & msk)] & 0xFF) << 8));
					pc += 2;
					dist = ((page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF) | ((page[(wk + 1) >> 13][wk + 1 - (wk & msk)] & 0xFF) << 8));
					pc += 2;
					length = ((page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF) | ((page[(wk + 1) >> 13][wk + 1 - (wk & msk)] & 0xFF) << 8));
					pc += 2;
					cycle = length * 6;
					do {
						pce._Wr6502(dist, pce._Rd6502(src));
						src++;
					} while (--length > 0);
				}
				break;

			case 0x14:
				J = ((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF));
				I = (ram[J] & 0xFF);
				NF = VF = I;
				I &= ~a;
				ZF = I;
				ram[J] = (byte)I;
				break;
				
			case 0x1C:
				J = (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF);
				J |= (((page[(wk = (pc + 1)) >> 13][wk - (wk & msk)] & 0xFF)) << 8);
				pc += 2;
				I = pce._Rd6502(J);
				NF = VF = I;
				I &= ~a;
				ZF = I;
				pce._Wr6502(J, I);
				break;

			case 0x04:
				J = ((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF));
				I = (ram[J] & 0xFF);
				NF = VF = I;
				I |= a;
				ZF = I;
				ram[J] = (byte)I;
				break;
				
			case 0x0C:
				J = (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF);
				J |= (((page[(wk = (pc + 1)) >> 13][wk - (wk & msk)] & 0xFF)) << 8);
				pc += 2;
				I = pce._Rd6502(J);
				NF = VF = I;
				I |= a;
				ZF = I;
				pce._Wr6502(J, I);
				break;

			case 0x83:
				I = (page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF);
				J = (J & 0xFF00) | (ram[((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF))] & 0xFF);
				NF = VF = ((J) & 0xFF);
				ZF = I & ((J) & 0xFF);
				break;
				
			case 0xA3:
				I = (page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF);
				J = (J & 0xFF00) | (ram[((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF) + x) & 0xFF] & 0xFF);
				NF = VF = ((J) & 0xFF);
				ZF = I & ((J) & 0xFF);
				break;
				
			case 0x93:
				I = (page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF);
				J = (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF);
				J |= (((page[(wk = (pc + 1)) >> 13][wk - (wk & msk)] & 0xFF)) << 8);
				pc += 2;
				J = (J & 0xFF00) | pce._Rd6502(J);
				NF = VF = ((J) & 0xFF);
				ZF = I & ((J) & 0xFF);
				break;
				
			case 0xB3:
				I = (page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF);
				J = (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF);
				J |= (((page[(wk = (pc + 1)) >> 13][wk - (wk & msk)] & 0xFF)) << 8);
				pc += 2;
				J += x;
				J = (J & 0xFF00) | pce._Rd6502(J);
				NF = VF = ((J) & 0xFF);
				ZF = I & ((J) & 0xFF);
				break;

			case 0x0F:
				if (((ram[((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF))] & 0xFF) & 0x01) != 0)
					pc++;
				else {
					pc += (byte) (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF) + 1;
					cycle += 2;
				}
				break;
				
			case 0x1F:
				if (((ram[((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF))] & 0xFF) & 0x02) != 0)
					pc++;
				else {
					pc += (byte) (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF) + 1;
					cycle += 2;
				}
				break;
				
			case 0x2F:
				if (((ram[((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF))] & 0xFF) & 0x04) != 0)
					pc++;
				else {
					pc += (byte) (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF) + 1;
					cycle += 2;
				}
				break;
				
			case 0x3F:
				if (((ram[((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF))] & 0xFF) & 0x08) != 0)
					pc++;
				else {
					pc += (byte) (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF) + 1;
					cycle += 2;
				}
				break;
				
			case 0x4F:
				if (((ram[((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF))] & 0xFF) & 0x10) != 0)
					pc++;
				else {
					pc += (byte) (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF) + 1;
					cycle += 2;
				}
				break;
				
			case 0x5F:
				if (((ram[((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF))] & 0xFF) & 0x20) != 0)
					pc++;
				else {
					pc += (byte) (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF) + 1;
					cycle += 2;
				}
				break;
				
			case 0x6F:
				if (((ram[((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF))] & 0xFF) & 0x40) != 0)
					pc++;
				else {
					pc += (byte) (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF) + 1;
					cycle += 2;
				}
				break;
				
			case 0x7F:
				if (((ram[((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF))] & 0xFF) & 0x80) != 0)
					pc++;
				else {
					pc += (byte) (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF) + 1;
					cycle += 2;
				}
				break;

			case 0x8F:
				if (((ram[((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF))] & 0xFF) & 0x01) != 0) {
					pc += (byte) (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF) + 1;
					cycle += 2;
				} else
					pc++;
				break;
				
			case 0x9F:
				if (((ram[((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF))] & 0xFF) & 0x02) != 0) {
					pc += (byte) (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF) + 1;
					cycle += 2;
				} else
					pc++;
				break;
				
			case 0xAF:
				if (((ram[((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF))] & 0xFF) & 0x04) != 0) {
					pc += (byte) (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF) + 1;
					cycle += 2;
				} else
					pc++;
				break;
				
			case 0xBF:
				if (((ram[((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF))] & 0xFF) & 0x08) != 0) {
					pc += (byte) (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF) + 1;
					cycle += 2;
				} else
					pc++;
				break;
				
			case 0xCF:
				if (((ram[((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF))] & 0xFF) & 0x10) != 0) {
					pc += (byte) (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF) + 1;
					cycle += 2;
				} else
					pc++;
				break;
				
			case 0xDF:
				if (((ram[((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF))] & 0xFF) & 0x20) != 0) {
					pc += (byte) (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF) + 1;
					cycle += 2;
				} else
					pc++;
				break;
				
			case 0xEF:
				if (((ram[((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF))] & 0xFF) & 0x40) != 0) {
					pc += (byte) (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF) + 1;
					cycle += 2;
				} else
					pc++;
				break;
				
			case 0xFF:
				if (((ram[((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF))] & 0xFF) & 0x80) != 0) {
					pc += (byte) (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF) + 1;
					cycle += 2;
				} else
					pc++;
				break;

			case 0x07:
				wk = page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF;
				ram[wk] = (byte)((ram[wk] & 0xFF) & ~0x01);
				break;
				
			case 0x17:
				wk = page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF;
				ram[wk] = (byte)((ram[wk] & 0xFF) & ~0x02);
				break;
				
			case 0x27:
				wk = page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF;
				ram[wk] = (byte)((ram[wk] & 0xFF) & ~0x04);
				break;
				
			case 0x37:
				wk = page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF;
				ram[wk] = (byte)((ram[wk] & 0xFF) & ~0x08);
				break;
				
			case 0x47:
				wk = page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF;
				ram[wk] = (byte)((ram[wk] & 0xFF) & ~0x10);
				break;
				
			case 0x57:
				wk = page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF;
				ram[wk] = (byte)((ram[wk] & 0xFF) & ~0x20);
				break;
				
			case 0x67:
				wk = page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF;
				ram[wk] = (byte)((ram[wk] & 0xFF) & ~0x40);
				break;
				
			case 0x77:
				wk = page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF;
				ram[wk] = (byte)((ram[wk] & 0xFF) & ~0x80);
				break;

			case 0x87:
				wk = page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF;
				ram[wk] = (byte)((ram[wk] & 0xFF) | 0x01);
				break;
				
			case 0x97:
				wk = page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF;
				ram[wk] = (byte)((ram[wk] & 0xFF) | 0x02);
				break;
				
			case 0xA7:
				wk = page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF;
				ram[wk] = (byte)((ram[wk] & 0xFF) | 0x04);
				break;
				
			case 0xB7:
				wk = page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF;
				ram[wk] = (byte)((ram[wk] & 0xFF) | 0x08);
				break;
				
			case 0xC7:
				wk = page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF;
				ram[wk] = (byte)((ram[wk] & 0xFF) | 0x10);
				break;
				
			case 0xD7:
				wk = page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF;
				ram[wk] = (byte)((ram[wk] & 0xFF) | 0x20);
				break;
				
			case 0xE7:
				wk = page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF;
				ram[wk] = (byte)((ram[wk] & 0xFF) | 0x40);
				break;
				
			case 0xF7:
				wk = page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF;
				ram[wk] = (byte)((ram[wk] & 0xFF) | 0x80);
				break;

			case 0x10:
				if ((NF & 0x80) != 0)
					pc++;
				else {
					pc += (byte) (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF) + 1;
					cycle += 2;
				}
				break;
				
			case 0x30:
				if ((NF & 0x80) != 0) {
					pc += (byte) (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF) + 1;
					cycle += 2;
				} else
					pc++;
				break;
				
			case 0xD0:
				if (ZF == 0)
					pc++;
				else {
					pc += (byte) (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF) + 1;
					cycle += 2;
				}
				break;
				
			case 0xF0:
				if (ZF == 0) {
					pc += (byte) (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF) + 1;
					cycle += 2;
				} else
					pc++;
				break;
				
			case 0x90:
				if ((p & 0x01) != 0)
					pc++;
				else {
					pc += (byte) (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF) + 1;
					cycle += 2;
				}
				break;
				
			case 0xB0:
				if ((p & 0x01) != 0) {
					pc += (byte) (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF) + 1;
					cycle += 2;
				} else
					pc++;
				break;
				
			case 0x50:
				if ((VF & 0x40) != 0)
					pc++;
				else {
					pc += (byte) (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF) + 1;
					cycle += 2;
				}
				break;
				
			case 0x70:
				if ((VF & 0x40) != 0) {
					pc += (byte) (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF) + 1;
					cycle += 2;
				} else
					pc++;
				break;

			case 0x40:
				I = p;
				s = (s + 1) & 0xFF;
				p = (ram[0x100 + s] & 0xFF);
				NF = VF = p;
				ZF = (((p & 0x02) != 0) ? 0 : 1);
				if ((IRequest != 0) && (I & 0x04) != 0 && (p & 0x04) == 0) {
					AfterCLI = 1;
					IBackup = ICount;
					ICount = 0;
				}
				s++;
				pc = (pc & 0xFF00) | (ram[0x100 + s] & 0xFF);
				s++;
				pc = (pc & 0xFF) | (((ram[0x100 + s] & 0xFF)) << 8);
				break;

			case 0x60:
				s++;
				pc = (pc & 0xFF00) | (ram[0x100 + s] & 0xFF);
				s++;
				pc = (pc & 0xFF) | (((ram[0x100 + s] & 0xFF)) << 8);
				pc++;
				break;

			case 0x20:
				K = (page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF);
				K |= (((page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF)) << 8);
				ram[0x100 + s] = (byte)(pc >> 8);
				s = (s - 1) & 0xFF;
				ram[0x100 + s] = (byte)(pc & 0xFF);
				s = (s - 1) & 0xFF;
				pc = K;
				break;

			case 0x4C:
				K = (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF);
				K |= (((page[(wk = (pc + 1)) >> 13][wk - (wk & msk)] & 0xFF)) << 8);
				pc += 2;
				pc = K;
				break;

			case 0x6C:
				K = (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF);
				K |= (((page[(wk = (pc + 1)) >> 13][wk - (wk & msk)] & 0xFF)) << 8);
				pc += 2;
				pc = (pc & 0xFF00)
						| (page[(wk = (K++)) >> 13][wk - (wk & msk)] & 0xFF);
				pc = (pc & 0xFF)
						| (((page[(wk = K) >> 13][wk - (wk & msk)] & 0xFF)) << 8);
				break;

			case 0x00:
				pc++;
				ram[0x100 + s] = (byte)(pc >> 8);
				s = (s - 1) & 0xFF;
				ram[0x100 + s] = (byte)(pc & 0xFF);
				s = (s - 1) & 0xFF;
				ram[0x100 + s] = (byte)(((p & ~0x20 | 0x10) & ~(0x80 | 0x40 | 0x02)) | (NF & 0x80) | (VF & 0x40) | ((ZF != 0) ? 0 : 0x02));
				s = (s - 1) & 0xFF;
				p = (p | 0x04) & ~0x08;
				pc = (page[(wk = (0xFFF6)) >> 13][wk - (wk & msk)] & 0xFF);
				pc |= (((page[(wk = (0xFFF6 + 1)) >> 13][wk
								- (wk & msk)] & 0xFF)) << 8);
				debug.TRACE("BRK instruction\n");
				break;

			case 0x58:
				if ((IRequest != 0) && (p & 0x04) != 0) {
					AfterCLI = 1;
					IBackup = ICount;
					ICount = 0;
				}
				p &= ~0x04;
				break;

			case 0x28:
				s = (s + 1) & 0xFF;
				I = (ram[0x100 + s] & 0xFF);
				NF = VF = I;
				ZF = (((I & 0x02) != 0) ? 0 : 1);
				if ((IRequest != 0) && ((I ^ p) & ~I & 0x04) != 0) {
					AfterCLI = 1;
					IBackup = ICount;
					ICount = 0;
				}
				p = I;
				break;

			case 0x08:
				ram[0x100 + s] = (byte)(((p & ~0x20 | 0x10) & ~(0x80 | 0x40 | 0x02))
						| (NF & 0x80) | (VF & 0x40) | ((ZF != 0) ? 0 : 0x02));
				s = (s - 1) & 0xFF;
				break;
				
			case 0x18:
				p &= ~0x01;
				break;
				
			case 0xB8:
				VF = 0;
				break;
				
			case 0xD8:
				p &= ~0x08;
				break;
				
			case 0x38:
				p |= 0x01;
				break;
				
			case 0xF8:
				p |= 0x08;
				break;
				
			case 0x78:
				p |= 0x04;
				break;
				
			case 0x48:
				ram[0x100 + s] = (byte)a;
				s = (s - 1) & 0xFF;
				break;
				
			case 0x68:
				s = (s + 1) & 0xFF;
				a = (ram[0x100 + s] & 0xFF);
				ZF = NF = a;
				break;
				
			case 0x98:
				a = y;
				ZF = NF = a;
				break;
				
			case 0xA8:
				y = a;
				ZF = NF = y;
				break;
				
			case 0xC8:
				y = (y + 1) & 0xFF;
				ZF = NF = y;
				break;
				
			case 0x88:
				y = (y - 1) & 0xFF;
				ZF = NF = y;
				break;
				
			case 0x8A:
				a = x;
				ZF = NF = a;
				break;
				
			case 0xAA:
				x = a;
				ZF = NF = x;
				break;
				
			case 0xE8:
				x = (x + 1) & 0xFF;
				ZF = NF = x;
				break;
				
			case 0xCA:
				x = (x - 1) & 0xFF;
				ZF = NF = x;
				break;
				
			case 0xEA:
				break;
				
			case 0x9A:
				s = x;
				break;
				
			case 0xBA:
				x = s;
				ZF = NF = x;
				break;

			case 0x24:
				I = (ram[((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF))] & 0xFF);
				NF = VF = I;
				ZF = I & a;
				break;
				
			case 0x2C:
				J = (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF);
				J |= (((page[(wk = (pc + 1)) >> 13][wk - (wk & msk)] & 0xFF)) << 8);
				pc += 2;
				I = pce._Rd6502(J);
				NF = VF = I;
				ZF = I & a;
				break;

			case 0x05:
				I = (ram[((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF))] & 0xFF);
				a |= I;
				ZF = NF = a;
				break;
				
			case 0x06:
				J = ((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF));
				I = (ram[J] & 0xFF);
				p &= ~0x01;
				p |= I >> 7;
				I = (I << 1) & 0xFF;
				ZF = NF = I;
				ram[J] = (byte)I;
				break;
				
			case 0x25:
				I = (ram[((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF))] & 0xFF);
				a &= I;
				ZF = NF = a;
				break;
				
			case 0x26:
				J = ((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF));
				I = (ram[J] & 0xFF);
				K = (K & 0xFF00) | ((I << 1) | (p & 0x01)) & 0xFF;
				p &= ~0x01;
				p |= I >> 7;
				I = (K & 0xFF);
				ZF = NF = I;
				ram[J] = (byte)I;
				break;
				
			case 0x45:
				I = (ram[((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF))] & 0xFF);
				a ^= I;
				ZF = NF = a;
				break;
				
			case 0x46:
				J = ((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF));
				I = (ram[J] & 0xFF);
				p &= ~0x01;
				p |= I & 0x01;
				I >>= 1;
				ZF = NF = I;
				ram[J] = (byte)I;
				break;
				
			case 0x65:
				I = (ram[((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF))] & 0xFF);
				if ((p & 0x08) != 0) {
					K = (a & 0x0F) + (I & 0x0F) + (p & 0x01);
					K |= (((a >> 4) + (I >> 4)) << 8);
					if ((K & 0xFF) > 9) {
						K = (K & 0xFF) + 6;
						K |= (((K >> 8) + 1) << 8);
					}
					if ((K >> 8) > 9)
						K = (K & 0xFF) | (((K >> 8) + 6) << 8);
					a = ((K & 0xFF) & 0x0F) | ((K >> 8) << 4);
					p = (p & ~0x01) | ((K >> 8) > 15 ? 0x01 : 0);
					ZF = NF = a;
					cycle++;
				} else {
					K = a + I + (p & 0x01);
					p &= ~0x01;
					p |= (((K >> 8) != 0) ? 0x01 : 0);
					VF = (~(a ^ I) & (a ^ (K & 0xFF))) >> 1;
					ZF = NF = (K & 0xFF);
					a = (K & 0xFF);
				}
				break;
				
			case 0x66:
				J = ((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF));
				I = (ram[J] & 0xFF);
				K = (K & 0xFF00) | (I >> 1) | (p << 7);
				p &= ~0x01;
				p |= I & 0x01;
				I = (K & 0xFF);
				ZF = NF = I;
				ram[J] = (byte)I;
				break;
				
			case 0x84:
				ram[((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF))] = (byte)y;
				break;
				
			case 0x85:
				ram[((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF))] = (byte)a;
				break;
				
			case 0x86:
				ram[((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF))] = (byte)x;
				break;
				
			case 0xA4:
				y = (ram[((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF))] & 0xFF);
				ZF = NF = y;
				break;
				
			case 0xA5:
				a = (ram[((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF))] & 0xFF);
				ZF = NF = a;
				break;
				
			case 0xA6:
				x = (ram[((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF))] & 0xFF);
				ZF = NF = x;
				break;
				
			case 0xC4:
				I = (ram[((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF))] & 0xFF);
				K = y - I;
				p &= ~0x01;
				p |= (((K >> 8) != 0) ? 0 : 0x01);
				ZF = NF = (K & 0xFF);
				break;
				
			case 0xC5:
				I = (ram[((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF))] & 0xFF);
				K = a - I;
				p &= ~0x01;
				p |= (((K >> 8) != 0) ? 0 : 0x01);
				ZF = NF = (K & 0xFF);
				break;
				
			case 0xC6:
				J = ((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF));
				I = (ram[J] & 0xFF);
				I = (I - 1) & 0xFF;
				ZF = NF = I;
				ram[J] = (byte)I;
				break;
				
			case 0xE4:
				I = (ram[((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF))] & 0xFF);
				K = x - I;
				p &= ~0x01;
				p |= (((K >> 8) != 0) ? 0 : 0x01);
				ZF = NF = (K & 0xFF);
				break;
				
			case 0xE5:
				I = (ram[((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF))] & 0xFF);
				if ((p & 0x08) != 0) {
					K = (K & 0xFF00) | (a & 0x0F) - (I & 0x0F) - (~p & 0x01);
					if (((K & 0xFF) & 0x10) != 0)
						K = (K & 0xFF00) | (K & 0xFF) - 6;
					K = (K & 0xFF) | (((a >> 4) - (I >> 4) - ((((K & 0xFF) & 0x10) == 0x10) ? 1 : 0)) << 8);
					if (((K >> 8) & 0x10) != 0)
						K = (K & 0xFF) | (((K >> 8) - 6) << 8);
					a = ((K & 0xFF) & 0x0F) | ((K >> 8) << 4);
					p = (p & ~0x01) | ((((K >> 8) & 0x10) != 0) ? 0 : 0x01);
					ZF = NF = a;
					cycle++;
				} else {
					K = a - I - (~p & 0x01);
					p &= ~0x01;
					p |= (((K >> 8) != 0) ? 0 : 0x01);
					VF = ((a ^ I) & (a ^ (K & 0xFF))) >> 1;
					ZF = NF = (K & 0xFF);
					a = (K & 0xFF);
				}
				break;
				
			case 0xE6:
				J = ((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF));
				I = (ram[J] & 0xFF);
				I = (I + 1) & 0xFF;
				ZF = NF = I;
				ram[J] = (byte)I;
				break;

			case 0x0D:
				J = (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF);
				J |= (((page[(wk = (pc + 1)) >> 13][wk - (wk & msk)] & 0xFF)) << 8);
				pc += 2;
				I = pce._Rd6502(J);
				a |= I;
				ZF = NF = a;
				break;
				
			case 0x0E:
				J = (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF);
				J |= (((page[(wk = (pc + 1)) >> 13][wk - (wk & msk)] & 0xFF)) << 8);
				pc += 2;
				I = pce._Rd6502(J);
				p &= ~0x01;
				p |= I >> 7;
				I = (I << 1) & 0xFF;
				ZF = NF = I;
				pce._Wr6502(J, I);
				break;
				
			case 0x2D:
				J = (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF);
				J |= (((page[(wk = (pc + 1)) >> 13][wk - (wk & msk)] & 0xFF)) << 8);
				pc += 2;
				I = pce._Rd6502(J);
				a &= I;
				ZF = NF = a;
				break;
				
			case 0x2E:
				J = (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF);
				J |= (((page[(wk = (pc + 1)) >> 13][wk - (wk & msk)] & 0xFF)) << 8);
				pc += 2;
				I = pce._Rd6502(J);
				K = (K & 0xFF00) | ((I << 1) | (p & 0x01)) & 0xFF;
				p &= ~0x01;
				p |= I >> 7;
				I = (K & 0xFF);
				ZF = NF = I;
				pce._Wr6502(J, I);
				break;
				
			case 0x4D:
				J = (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF);
				J |= (((page[(wk = (pc + 1)) >> 13][wk - (wk & msk)] & 0xFF)) << 8);
				pc += 2;
				I = pce._Rd6502(J);
				a ^= I;
				ZF = NF = a;
				break;
				
			case 0x4E:
				J = (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF);
				J |= (((page[(wk = (pc + 1)) >> 13][wk - (wk & msk)] & 0xFF)) << 8);
				pc += 2;
				I = pce._Rd6502(J);
				p &= ~0x01;
				p |= I & 0x01;
				I >>= 1;
				ZF = NF = I;
				pce._Wr6502(J, I);
				break;
				
			case 0x6D:
				J = (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF);
				J |= (((page[(wk = (pc + 1)) >> 13][wk - (wk & msk)] & 0xFF)) << 8);
				pc += 2;
				I = pce._Rd6502(J);
				if ((p & 0x08) != 0) {
					K = (a & 0x0F) + (I & 0x0F) + (p & 0x01);
					K |= (((a >> 4) + (I >> 4)) << 8);
					if ((K & 0xFF) > 9) {
						K = (K & 0xFF) + 6;
						K |= (((K >> 8) + 1) << 8);
					}
					if ((K >> 8) > 9)
						K = (K & 0xFF) | (((K >> 8) + 6) << 8);
					a = ((K & 0xFF) & 0x0F) | ((K >> 8) << 4);
					p = (p & ~0x01) | ((K >> 8) > 15 ? 0x01 : 0);
					ZF = NF = a;
					cycle++;
				} else {
					K = a + I + (p & 0x01);
					p &= ~0x01;
					p |= (((K >> 8) != 0) ? 0x01 : 0);
					VF = (~(a ^ I) & (a ^ (K & 0xFF))) >> 1;
					ZF = NF = (K & 0xFF);
					a = (K & 0xFF);
				}
				break;
				
			case 0x6E:
				J = (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF);
				J |= (((page[(wk = (pc + 1)) >> 13][wk - (wk & msk)] & 0xFF)) << 8);
				pc += 2;
				I = pce._Rd6502(J);
				K = (K & 0xFF00) | (I >> 1) | (p << 7);
				p &= ~0x01;
				p |= I & 0x01;
				I = (K & 0xFF);
				ZF = NF = I;
				pce._Wr6502(J, I);
				break;
				
			case 0x8C:
				J = (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF);
				J |= (((page[(wk = (pc + 1)) >> 13][wk - (wk & msk)] & 0xFF)) << 8);
				pc += 2;
				pce._Wr6502(J, y);
				break;
				
			case 0x8D:
				J = (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF);
				J |= (((page[(wk = (pc + 1)) >> 13][wk - (wk & msk)] & 0xFF)) << 8);
				pc += 2;
				pce._Wr6502(J, a);
				break;
				
			case 0x8E:
				J = (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF);
				J |= (((page[(wk = (pc + 1)) >> 13][wk - (wk & msk)] & 0xFF)) << 8);
				pc += 2;
				pce._Wr6502(J, x);
				break;
				
			case 0xAC:
				J = (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF);
				J |= (((page[(wk = (pc + 1)) >> 13][wk - (wk & msk)] & 0xFF)) << 8);
				pc += 2;
				y = pce._Rd6502(J);
				ZF = NF = y;
				break;
				
			case 0xAD:
				J = (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF);
				J |= (((page[(wk = (pc + 1)) >> 13][wk - (wk & msk)] & 0xFF)) << 8);
				pc += 2;
				a = pce._Rd6502(J);
				ZF = NF = a;
				break;
				
			case 0xAE:
				J = (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF);
				J |= (((page[(wk = (pc + 1)) >> 13][wk - (wk & msk)] & 0xFF)) << 8);
				pc += 2;
				x = pce._Rd6502(J);
				ZF = NF = x;
				break;
				
			case 0xCC:
				J = (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF);
				J |= (((page[(wk = (pc + 1)) >> 13][wk - (wk & msk)] & 0xFF)) << 8);
				pc += 2;
				I = pce._Rd6502(J);
				K = y - I;
				p &= ~0x01;
				p |= (((K >> 8) != 0) ? 0 : 0x01);
				ZF = NF = (K & 0xFF);
				break;
				
			case 0xCD:
				J = (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF);
				J |= (((page[(wk = (pc + 1)) >> 13][wk - (wk & msk)] & 0xFF)) << 8);
				pc += 2;
				I = pce._Rd6502(J);
				K = a - I;
				p &= ~0x01;
				p |= (((K >> 8) != 0) ? 0 : 0x01);
				ZF = NF = (K & 0xFF);
				break;
				
			case 0xCE:
				J = (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF);
				J |= (((page[(wk = (pc + 1)) >> 13][wk - (wk & msk)] & 0xFF)) << 8);
				pc += 2;
				I = pce._Rd6502(J);
				I = (I - 1) & 0xFF;
				ZF = NF = I;
				pce._Wr6502(J, I);
				break;
				
			case 0xEC:
				J = (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF);
				J |= (((page[(wk = (pc + 1)) >> 13][wk - (wk & msk)] & 0xFF)) << 8);
				pc += 2;
				I = pce._Rd6502(J);
				K = x - I;
				p &= ~0x01;
				p |= (((K >> 8) != 0) ? 0 : 0x01);
				ZF = NF = (K & 0xFF);
				break;
				
			case 0xED:
				J = (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF);
				J |= (((page[(wk = (pc + 1)) >> 13][wk - (wk & msk)] & 0xFF)) << 8);
				pc += 2;
				I = pce._Rd6502(J);
				if ((p & 0x08) != 0) {
					K = (K & 0xFF00) | (a & 0x0F) - (I & 0x0F) - (~p & 0x01);
					if (((K & 0xFF) & 0x10) != 0)
						K = (K & 0xFF00) | (K & 0xFF) - 6;
					K = (K & 0xFF)
							| (((a >> 4) - (I >> 4) - ((((K & 0xFF) & 0x10) == 0x10) ? 1 : 0)) << 8);
					if (((K >> 8) & 0x10) != 0)
						K = (K & 0xFF) | (((K >> 8) - 6) << 8);
					a = ((K & 0xFF) & 0x0F) | ((K >> 8) << 4);
					p = (p & ~0x01) | ((((K >> 8) & 0x10) != 0) ? 0 : 0x01);
					ZF = NF = a;
					cycle++;
				} else {
					K = a - I - (~p & 0x01);
					p &= ~0x01;
					p |= (((K >> 8) != 0) ? 0 : 0x01);
					VF = ((a ^ I) & (a ^ (K & 0xFF))) >> 1;
					ZF = NF = (K & 0xFF);
					a = (K & 0xFF);
				}
				break;
				
			case 0xEE:
				J = (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF);
				J |= (((page[(wk = (pc + 1)) >> 13][wk - (wk & msk)] & 0xFF)) << 8);
				pc += 2;
				I = pce._Rd6502(J);
				I = (I + 1) & 0xFF;
				ZF = NF = I;
				pce._Wr6502(J, I);
				break;

			case 0x09:
				I = (page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF);
				a |= I;
				ZF = NF = a;
				break;
				
			case 0x29:
				I = (page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF);
				a &= I;
				ZF = NF = a;
				break;
				
			case 0x49:
				I = (page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF);
				a ^= I;
				ZF = NF = a;
				break;
				
			case 0x69:
				I = (page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF);
				if ((p & 0x08) != 0) {
					K = (a & 0x0F) + (I & 0x0F) + (p & 0x01);
					K |= (((a >> 4) + (I >> 4)) << 8);
					if ((K & 0xFF) > 9) {
						K = (K & 0xFF) + 6;
						K |= (((K >> 8) + 1) << 8);
					}
					if ((K >> 8) > 9)
						K = (K & 0xFF) | (((K >> 8) + 6) << 8);
					a = ((K & 0xFF) & 0x0F) | ((K >> 8) << 4);
					p = (p & ~0x01) | ((K >> 8) > 15 ? 0x01 : 0);
					ZF = NF = a;
					cycle++;
				} else {
					K = a + I + (p & 0x01);
					p &= ~0x01;
					p |= (((K >> 8) != 0) ? 0x01 : 0);
					VF = (~(a ^ I) & (a ^ (K & 0xFF))) >> 1;
					ZF = NF = (K & 0xFF);
					a = (K & 0xFF);
				}
				break;
				
			case 0xA0:
				y = (page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF);
				ZF = NF = y;
				break;
				
			case 0xA2:
				x = (page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF);
				ZF = NF = x;
				break;
				
			case 0xA9:
				a = (page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF);
				ZF = NF = a;
				break;
				
			case 0xC0:
				I = (page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF);
				K = y - I;
				p &= ~0x01;
				p |= (((K >> 8) != 0) ? 0 : 0x01);
				ZF = NF = (K & 0xFF);
				break;
				
			case 0xC9:
				I = (page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF);
				K = a - I;
				p &= ~0x01;
				p |= (((K >> 8) != 0) ? 0 : 0x01);
				ZF = NF = (K & 0xFF);
				break;
				
			case 0xE0:
				I = (page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF);
				K = x - I;
				p &= ~0x01;
				p |= (((K >> 8) != 0) ? 0 : 0x01);
				ZF = NF = (K & 0xFF);
				break;
				
			case 0xE9:
				I = (page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF);
				if ((p & 0x08) != 0) {
					K = (K & 0xFF00) | (a & 0x0F) - (I & 0x0F) - (~p & 0x01);
					if (((K & 0xFF) & 0x10) != 0)
						K = (K & 0xFF00) | (K & 0xFF) - 6;
					K = (K & 0xFF)
							| (((a >> 4) - (I >> 4) - ((((K & 0xFF) & 0x10) == 0x10) ? 1
									: 0)) << 8);
					if (((K >> 8) & 0x10) != 0)
						K = (K & 0xFF) | (((K >> 8) - 6) << 8);
					a = ((K & 0xFF) & 0x0F) | ((K >> 8) << 4);
					p = (p & ~0x01) | ((((K >> 8) & 0x10) != 0) ? 0 : 0x01);
					ZF = NF = a;
					cycle++;
				} else {
					K = a - I - (~p & 0x01);
					p &= ~0x01;
					p |= (((K >> 8) != 0) ? 0 : 0x01);
					VF = ((a ^ I) & (a ^ (K & 0xFF))) >> 1;
					ZF = NF = (K & 0xFF);
					a = (K & 0xFF);
				}
				break;

			case 0x15:
				I = (ram[((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF) + x) & 0xFF] & 0xFF);
				a |= I;
				ZF = NF = a;
				break;
				
			case 0x16:
				J = ((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF) + x) & 0xFF;
				I = (ram[J] & 0xFF);
				p &= ~0x01;
				p |= I >> 7;
				I = (I << 1) & 0xFF;
				ZF = NF = I;
				ram[J] = (byte)I;
				break;
				
			case 0x35:
				I = (ram[((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF) + x) & 0xFF] & 0xFF);
				a &= I;
				ZF = NF = a;
				break;
				
			case 0x36:
				J = ((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF) + x) & 0xFF;
				I = (ram[J] & 0xFF);
				K = (K & 0xFF00) | ((I << 1) | (p & 0x01)) & 0xFF;
				p &= ~0x01;
				p |= I >> 7;
				I = (K & 0xFF);
				ZF = NF = I;
				ram[J] = (byte)I;
				break;
				
			case 0x55:
				I = (ram[((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF) + x) & 0xFF] & 0xFF);
				a ^= I;
				ZF = NF = a;
				break;
				
			case 0x56:
				J = ((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF) + x) & 0xFF;
				I = (ram[J] & 0xFF);
				p &= ~0x01;
				p |= I & 0x01;
				I >>= 1;
				ZF = NF = I;
				ram[J] = (byte)I;
				break;
				
			case 0x75:
				I = (ram[((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF) + x)] & 0xFF) & 0xFF;
				if ((p & 0x08) != 0) {
					K = (a & 0x0F) + (I & 0x0F) + (p & 0x01);
					K |= (((a >> 4) + (I >> 4)) << 8);
					if ((K & 0xFF) > 9) {
						K = (K & 0xFF) + 6;
						K |= (((K >> 8) + 1) << 8);
					}
					if ((K >> 8) > 9)
						K = (K & 0xFF) | (((K >> 8) + 6) << 8);
					a = ((K & 0xFF) & 0x0F) | ((K >> 8) << 4);
					p = (p & ~0x01) | ((K >> 8) > 15 ? 0x01 : 0);
					ZF = NF = a;
					cycle++;
				} else {
					K = a + I + (p & 0x01);
					p &= ~0x01;
					p |= (((K >> 8) != 0) ? 0x01 : 0);
					VF = (~(a ^ I) & (a ^ (K & 0xFF))) >> 1;
					ZF = NF = (K & 0xFF);
					a = (K & 0xFF);
				}
				break;
				
			case 0x76:
				J = ((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF) + x) & 0xFF;
				I = (ram[J] & 0xFF);
				K = (K & 0xFF00) | (I >> 1) | (p << 7);
				p &= ~0x01;
				p |= I & 0x01;
				I = (K & 0xFF);
				ZF = NF = I;
				ram[J] = (byte)I;
				break;
				
			case 0x94:
				ram[((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF) + x) & 0xFF] = (byte)y;
				break;
				
			case 0x95:
				ram[((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF) + x) & 0xFF] = (byte)a;
				break;
				
			case 0x96:
				ram[((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF) + y) & 0xFF] = (byte)x;
				break;
				
			case 0xB4:
				y = (ram[((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF) + x) & 0xFF] & 0xFF);
				ZF = NF = y;
				break;
				
			case 0xB5:
				a = (ram[((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF) + x) & 0xFF] & 0xFF);
				ZF = NF = a;
				break;
				
			case 0xB6:
				x = (ram[((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF) + y) & 0xFF] & 0xFF);
				ZF = NF = x;
				break;
				
			case 0xD5:
				I = (ram[((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF) + x) & 0xFF] & 0xFF);
				K = a - I;
				p &= ~0x01;
				p |= (((K >> 8) != 0) ? 0 : 0x01);
				ZF = NF = (K & 0xFF);
				break;
				
			case 0xD6:
				J = ((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF) + x) & 0xFF;
				I = (ram[J] & 0xFF);
				I = (I - 1) & 0xFF;
				ZF = NF = I;
				ram[J] = (byte)I;
				break;
				
			case 0xF5:
				I = (ram[((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF) + x) & 0xFF] & 0xFF);
				if ((p & 0x08) != 0) {
					K = (K & 0xFF00) | (a & 0x0F) - (I & 0x0F) - (~p & 0x01);
					if (((K & 0xFF) & 0x10) != 0)
						K = (K & 0xFF00) | (K & 0xFF) - 6;
					K = (K & 0xFF)
							| (((a >> 4) - (I >> 4) - ((((K & 0xFF) & 0x10) == 0x10) ? 1
									: 0)) << 8);
					if (((K >> 8) & 0x10) != 0)
						K = (K & 0xFF) | (((K >> 8) - 6) << 8);
					a = ((K & 0xFF) & 0x0F) | ((K >> 8) << 4);
					p = (p & ~0x01) | ((((K >> 8) & 0x10) != 0) ? 0 : 0x01);
					ZF = NF = a;
					cycle++;
				} else {
					K = a - I - (~p & 0x01);
					p &= ~0x01;
					p |= (((K >> 8) != 0) ? 0 : 0x01);
					VF = ((a ^ I) & (a ^ (K & 0xFF))) >> 1;
					ZF = NF = (K & 0xFF);
					a = (K & 0xFF);
				}
				break;
				
			case 0xF6:
				J = ((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF) + x) & 0xFF;
				I = (ram[J] & 0xFF);
				I = (I + 1) & 0xFF;
				ZF = NF = I;
				ram[J] = (byte)I;
				break;

			case 0x19:
				J = (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF);
				J |= (((page[(wk = (pc + 1)) >> 13][wk - (wk & msk)] & 0xFF)) << 8);
				pc += 2;
				J += y;
				I = pce._Rd6502(J);
				a |= I;
				ZF = NF = a;
				break;
				
			case 0x1D:
				J = (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF);
				J |= (((page[(wk = (pc + 1)) >> 13][wk - (wk & msk)] & 0xFF)) << 8);
				pc += 2;
				J += x;
				I = pce._Rd6502(J);
				a |= I;
				ZF = NF = a;
				break;
				
			case 0x1E:
				J = (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF);
				J |= (((page[(wk = (pc + 1)) >> 13][wk - (wk & msk)] & 0xFF)) << 8);
				pc += 2;
				J += x;
				I = pce._Rd6502(J);
				p &= ~0x01;
				p |= I >> 7;
				I = (I << 1) & 0xFF;
				ZF = NF = I;
				pce._Wr6502(J, I);
				break;
				
			case 0x39:
				J = (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF);
				J |= (((page[(wk = (pc + 1)) >> 13][wk - (wk & msk)] & 0xFF)) << 8);
				pc += 2;
				J += y;
				I = pce._Rd6502(J);
				a &= I;
				ZF = NF = a;
				break;
				
			case 0x3D:
				J = (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF);
				J |= (((page[(wk = (pc + 1)) >> 13][wk - (wk & msk)] & 0xFF)) << 8);
				pc += 2;
				J += x;
				I = pce._Rd6502(J);
				a &= I;
				ZF = NF = a;
				break;
				
			case 0x3E:
				J = (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF);
				J |= (((page[(wk = (pc + 1)) >> 13][wk - (wk & msk)] & 0xFF)) << 8);
				pc += 2;
				J += x;
				I = pce._Rd6502(J);
				K = (K & 0xFF00) | ((I << 1) | (p & 0x01)) & 0xFF;
				p &= ~0x01;
				p |= I >> 7;
				I = (K & 0xFF);
				ZF = NF = I;
				pce._Wr6502(J, I);
				break;
				
			case 0x59:
				J = (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF);
				J |= (((page[(wk = (pc + 1)) >> 13][wk - (wk & msk)] & 0xFF)) << 8);
				pc += 2;
				J += y;
				I = pce._Rd6502(J);
				a ^= I;
				ZF = NF = a;
				break;
				
			case 0x5D:
				J = (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF);
				J |= (((page[(wk = (pc + 1)) >> 13][wk - (wk & msk)] & 0xFF)) << 8);
				pc += 2;
				J += x;
				I = pce._Rd6502(J);
				a ^= I;
				ZF = NF = a;
				break;
				
			case 0x5E:
				J = (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF);
				J |= (((page[(wk = (pc + 1)) >> 13][wk - (wk & msk)] & 0xFF)) << 8);
				pc += 2;
				J += x;
				I = pce._Rd6502(J);
				p &= ~0x01;
				p |= I & 0x01;
				I >>= 1;
				ZF = NF = I;
				pce._Wr6502(J, I);
				break;
				
			case 0x79:
				J = (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF);
				J |= (((page[(wk = (pc + 1)) >> 13][wk - (wk & msk)] & 0xFF)) << 8);
				pc += 2;
				J += y;
				I = pce._Rd6502(J);
				if ((p & 0x08) != 0) {
					K = (a & 0x0F) + (I & 0x0F) + (p & 0x01);
					K |= (((a >> 4) + (I >> 4)) << 8);
					if ((K & 0xFF) > 9) {
						K = (K & 0xFF) + 6;
						K |= (((K >> 8) + 1) << 8);
					}
					if ((K >> 8) > 9)
						K = (K & 0xFF) | (((K >> 8) + 6) << 8);
					a = ((K & 0xFF) & 0x0F) | ((K >> 8) << 4);
					p = (p & ~0x01) | ((K >> 8) > 15 ? 0x01 : 0);
					ZF = NF = a;
					cycle++;
				} else {
					K = a + I + (p & 0x01);
					p &= ~0x01;
					p |= (((K >> 8) != 0) ? 0x01 : 0);
					VF = (~(a ^ I) & (a ^ (K & 0xFF))) >> 1;
					ZF = NF = (K & 0xFF);
					a = (K & 0xFF);
				}
				break;
				
			case 0x7D:
				J = (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF);
				J |= (((page[(wk = (pc + 1)) >> 13][wk - (wk & msk)] & 0xFF)) << 8);
				pc += 2;
				J += x;
				I = pce._Rd6502(J);
				if ((p & 0x08) != 0) {
					K = (a & 0x0F) + (I & 0x0F) + (p & 0x01);
					K |= (((a >> 4) + (I >> 4)) << 8);
					if ((K & 0xFF) > 9) {
						K = (K & 0xFF) + 6;
						K |= (((K >> 8) + 1) << 8);
					}
					if ((K >> 8) > 9)
						K = (K & 0xFF) | (((K >> 8) + 6) << 8);
					a = ((K & 0xFF) & 0x0F) | ((K >> 8) << 4);
					p = (p & ~0x01) | ((K >> 8) > 15 ? 0x01 : 0);
					ZF = NF = a;
					cycle++;
				} else {
					K = a + I + (p & 0x01);
					p &= ~0x01;
					p |= (((K >> 8) != 0) ? 0x01 : 0);
					VF = (~(a ^ I) & (a ^ (K & 0xFF))) >> 1;
					ZF = NF = (K & 0xFF);
					a = (K & 0xFF);
				}
				break;
				
			case 0x7E:
				J = (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF);
				J |= (((page[(wk = (pc + 1)) >> 13][wk - (wk & msk)] & 0xFF)) << 8);
				pc += 2;
				J += x;
				I = pce._Rd6502(J);
				K = (K & 0xFF00) | (I >> 1) | (p << 7);
				p &= ~0x01;
				p |= I & 0x01;
				I = (K & 0xFF);
				ZF = NF = I;
				pce._Wr6502(J, I);
				break;
				
			case 0x99:
				J = (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF);
				J |= (((page[(wk = (pc + 1)) >> 13][wk - (wk & msk)] & 0xFF)) << 8);
				pc += 2;
				J += y;
				pce._Wr6502(J, a);
				break;
				
			case 0x9D:
				J = (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF);
				J |= (((page[(wk = (pc + 1)) >> 13][wk - (wk & msk)] & 0xFF)) << 8);
				pc += 2;
				J += x;
				pce._Wr6502(J, a);
				break;
				
			case 0xB9:
				J = (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF);
				J |= (((page[(wk = (pc + 1)) >> 13][wk - (wk & msk)] & 0xFF)) << 8);
				pc += 2;
				J += y;
				a = pce._Rd6502(J);
				ZF = NF = a;
				break;
				
			case 0xBC:
				J = (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF);
				J |= (((page[(wk = (pc + 1)) >> 13][wk - (wk & msk)] & 0xFF)) << 8);
				pc += 2;
				J += x;
				y = pce._Rd6502(J);
				ZF = NF = y;
				break;
				
			case 0xBD:
				J = (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF);
				J |= (((page[(wk = (pc + 1)) >> 13][wk - (wk & msk)] & 0xFF)) << 8);
				pc += 2;
				J += x;
				a = pce._Rd6502(J);
				ZF = NF = a;
				break;
				
			case 0xBE:
				J = (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF);
				J |= (((page[(wk = (pc + 1)) >> 13][wk - (wk & msk)] & 0xFF)) << 8);
				pc += 2;
				J += y;
				x = pce._Rd6502(J);
				ZF = NF = x;
				break;
				
			case 0xD9:
				J = (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF);
				J |= (((page[(wk = (pc + 1)) >> 13][wk - (wk & msk)] & 0xFF)) << 8);
				pc += 2;
				J += y;
				I = pce._Rd6502(J);
				K = a - I;
				p &= ~0x01;
				p |= (((K >> 8) != 0) ? 0 : 0x01);
				ZF = NF = (K & 0xFF);
				break;
				
			case 0xDD:
				J = (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF);
				J |= (((page[(wk = (pc + 1)) >> 13][wk - (wk & msk)] & 0xFF)) << 8);
				pc += 2;
				J += x;
				I = pce._Rd6502(J);
				K = a - I;
				p &= ~0x01;
				p |= (((K >> 8) != 0) ? 0 : 0x01);
				ZF = NF = (K & 0xFF);
				break;
				
			case 0xDE:
				J = (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF);
				J |= (((page[(wk = (pc + 1)) >> 13][wk - (wk & msk)] & 0xFF)) << 8);
				pc += 2;
				J += x;
				I = pce._Rd6502(J);
				I = (I - 1) & 0xFF;
				ZF = NF = I;
				pce._Wr6502(J, I);
				break;
				
			case 0xF9:
				J = (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF);
				J |= (((page[(wk = (pc + 1)) >> 13][wk - (wk & msk)] & 0xFF)) << 8);
				pc += 2;
				J += y;
				I = pce._Rd6502(J);
				if ((p & 0x08) != 0) {
					K = (K & 0xFF00) | (a & 0x0F) - (I & 0x0F) - (~p & 0x01);
					if (((K & 0xFF) & 0x10) != 0)
						K = (K & 0xFF00) | (K & 0xFF) - 6;
					K = (K & 0xFF)
							| (((a >> 4) - (I >> 4) - ((((K & 0xFF) & 0x10) == 0x10) ? 1
									: 0)) << 8);
					if (((K >> 8) & 0x10) != 0)
						K = (K & 0xFF) | (((K >> 8) - 6) << 8);
					a = ((K & 0xFF) & 0x0F) | ((K >> 8) << 4);
					p = (p & ~0x01) | ((((K >> 8) & 0x10) != 0) ? 0 : 0x01);
					ZF = NF = a;
					cycle++;
				} else {
					K = a - I - (~p & 0x01);
					p &= ~0x01;
					p |= (((K >> 8) != 0) ? 0 : 0x01);
					VF = ((a ^ I) & (a ^ (K & 0xFF))) >> 1;
					ZF = NF = (K & 0xFF);
					a = (K & 0xFF);
				}
				break;
				
			case 0xFD:
				J = (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF);
				J |= (((page[(wk = (pc + 1)) >> 13][wk - (wk & msk)] & 0xFF)) << 8);
				pc += 2;
				J += x;
				I = pce._Rd6502(J);
				if ((p & 0x08) != 0) {
					K = (K & 0xFF00) | (a & 0x0F) - (I & 0x0F) - (~p & 0x01);
					if (((K & 0xFF) & 0x10) != 0)
						K = (K & 0xFF00) | (K & 0xFF) - 6;
					K = (K & 0xFF) | (((a >> 4) - (I >> 4) - ((((K & 0xFF) & 0x10) == 0x10) ? 1 : 0)) << 8);
					if (((K >> 8) & 0x10) != 0)
						K = (K & 0xFF) | (((K >> 8) - 6) << 8);
					a = ((K & 0xFF) & 0x0F) | ((K >> 8) << 4);
					p = (p & ~0x01) | ((((K >> 8) & 0x10) != 0) ? 0 : 0x01);
					ZF = NF = a;
					cycle++;
				} else {
					K = a - I - (~p & 0x01);
					p &= ~0x01;
					p |= (((K >> 8) != 0) ? 0 : 0x01);
					VF = ((a ^ I) & (a ^ (K & 0xFF))) >> 1;
					ZF = NF = (K & 0xFF);
					a = (K & 0xFF);
				}
				break;
				
			case 0xFE:
				J = (page[(wk = pc) >> 13][wk - (wk & msk)] & 0xFF);
				J |= (((page[(wk = (pc + 1)) >> 13][wk - (wk & msk)] & 0xFF)) << 8);
				pc += 2;
				J += x;
				I = pce._Rd6502(J);
				I = (I + 1) & 0xFF;
				ZF = NF = I;
				pce._Wr6502(J, I);
				break;

			case 0x01:
				K = ((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF) + x) & 0xFF;
				J = (ram[K] & 0xFF);
				J |= (((ram[K + 1] & 0xFF)) << 8);
				I = pce._Rd6502(J);
				a |= I;
				ZF = NF = a;
				break;
				
			case 0x11:
				K = ((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF));
				J = (ram[K] & 0xFF);
				J |= (((ram[K + 1] & 0xFF)) << 8);
				J += y;
				I = pce._Rd6502(J);
				a |= I;
				ZF = NF = a;
				break;
				
			case 0x21:
				K = ((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF) + x) & 0xFF;
				J = (ram[K] & 0xFF);
				J |= (((ram[K + 1] & 0xFF)) << 8);
				I = pce._Rd6502(J);
				a &= I;
				ZF = NF = a;
				break;
				
			case 0x31:
				K = ((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF));
				J = (ram[K] & 0xFF);
				J |= (((ram[K + 1] & 0xFF)) << 8);
				J += y;
				I = pce._Rd6502(J);
				a &= I;
				ZF = NF = a;
				break;
				
			case 0x41:
				K = ((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF) + x) & 0xFF;
				J = (ram[K] & 0xFF);
				J |= (((ram[K + 1] & 0xFF)) << 8);
				I = pce._Rd6502(J);
				a ^= I;
				ZF = NF = a;
				break;
				
			case 0x51:
				K = ((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF));
				J = (ram[K] & 0xFF);
				J |= (((ram[K + 1] & 0xFF)) << 8);
				J += y;
				I = pce._Rd6502(J);
				a ^= I;
				ZF = NF = a;
				break;
				
			case 0x61:
				K = ((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF) + x) & 0xFF;
				J = (ram[K] & 0xFF);
				J |= (((ram[K + 1] & 0xFF)) << 8);
				I = pce._Rd6502(J);
				if ((p & 0x08) != 0) {
					K = (a & 0x0F) + (I & 0x0F) + (p & 0x01);
					K |= (((a >> 4) + (I >> 4)) << 8);
					if ((K & 0xFF) > 9) {
						K = (K & 0xFF) + 6;
						K |= (((K >> 8) + 1) << 8);
					}
					if ((K >> 8) > 9)
						K = (K & 0xFF) | (((K >> 8) + 6) << 8);
					a = ((K & 0xFF) & 0x0F) | ((K >> 8) << 4);
					p = (p & ~0x01) | ((K >> 8) > 15 ? 0x01 : 0);
					ZF = NF = a;
					cycle++;
				} else {
					K = a + I + (p & 0x01);
					p &= ~0x01;
					p |= (((K >> 8) != 0) ? 0x01 : 0);
					VF = (~(a ^ I) & (a ^ (K & 0xFF))) >> 1;
					ZF = NF = (K & 0xFF);
					a = (K & 0xFF);
				}
				break;
				
			case 0x71:
				K = ((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF));
				J = (ram[K] & 0xFF);
				J |= (((ram[K + 1] & 0xFF)) << 8);
				J += y;
				I = pce._Rd6502(J);
				if ((p & 0x08) != 0) {
					K = (a & 0x0F) + (I & 0x0F) + (p & 0x01);
					K |= (((a >> 4) + (I >> 4)) << 8);
					if ((K & 0xFF) > 9) {
						K = (K & 0xFF) + 6;
						K |= (((K >> 8) + 1) << 8);
					}
					if ((K >> 8) > 9)
						K = (K & 0xFF) | (((K >> 8) + 6) << 8);
					a = ((K & 0xFF) & 0x0F) | ((K >> 8) << 4);
					p = (p & ~0x01) | ((K >> 8) > 15 ? 0x01 : 0);
					ZF = NF = a;
					cycle++;
				} else {
					K = a + I + (p & 0x01);
					p &= ~0x01;
					p |= (((K >> 8) != 0) ? 0x01 : 0);
					VF = (~(a ^ I) & (a ^ (K & 0xFF))) >> 1;
					ZF = NF = (K & 0xFF);
					a = (K & 0xFF);
				}
				break;
				
			case 0x81:
				K = ((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF) + x) & 0xFF;
				J = (ram[K] & 0xFF);
				J |= (((ram[K + 1] & 0xFF)) << 8);
				pce._Wr6502(J, a);
				break;
				
			case 0x91:
				K = ((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF));
				J = (ram[K] & 0xFF);
				J |= (((ram[K + 1] & 0xFF)) << 8);
				J += y;
				pce._Wr6502(J, a);
				break;
				
			case 0xA1:
				K = ((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF) + x) & 0xFF;
				J = (ram[K] & 0xFF);
				J |= (((ram[K + 1] & 0xFF)) << 8);
				a = pce._Rd6502(J);
				ZF = NF = a;
				break;
				
			case 0xB1:
				K = ((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF));
				J = (ram[K] & 0xFF);
				J |= (((ram[K + 1] & 0xFF)) << 8);
				J += y;
				a = pce._Rd6502(J);
				ZF = NF = a;
				break;
				
			case 0xC1:
				K = ((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF) + x) & 0xFF;
				J = (ram[K] & 0xFF);
				J |= (((ram[K + 1] & 0xFF)) << 8);
				I = pce._Rd6502(J);
				K = a - I;
				p &= ~0x01;
				p |= (((K >> 8) != 0) ? 0 : 0x01);
				ZF = NF = (K & 0xFF);
				break;
				
			case 0xD1:
				K = ((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF));
				J = (ram[K] & 0xFF);
				J |= (((ram[K + 1] & 0xFF)) << 8);
				J += y;
				I = pce._Rd6502(J);
				K = a - I;
				p &= ~0x01;
				p |= (((K >> 8) != 0) ? 0 : 0x01);
				ZF = NF = (K & 0xFF);
				break;
				
			case 0xE1:
				K = ((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF) + x) & 0xFF;
				J = (ram[K] & 0xFF);
				J |= (((ram[K + 1] & 0xFF)) << 8);
				I = pce._Rd6502(J);
				if ((p & 0x08) != 0) {
					K = (K & 0xFF00) | (a & 0x0F) - (I & 0x0F) - (~p & 0x01);
					if (((K & 0xFF) & 0x10) != 0)
						K = (K & 0xFF00) | (K & 0xFF) - 6;
					K = (K & 0xFF) | (((a >> 4) - (I >> 4) - ((((K & 0xFF) & 0x10) == 0x10) ? 1 : 0)) << 8);
					if (((K >> 8) & 0x10) != 0)
						K = (K & 0xFF) | (((K >> 8) - 6) << 8);
					a = ((K & 0xFF) & 0x0F) | ((K >> 8) << 4);
					p = (p & ~0x01) | ((((K >> 8) & 0x10) != 0) ? 0 : 0x01);
					ZF = NF = a;
					cycle++;
				} else {
					K = a - I - (~p & 0x01);
					p &= ~0x01;
					p |= (((K >> 8) != 0) ? 0 : 0x01);
					VF = ((a ^ I) & (a ^ (K & 0xFF))) >> 1;
					ZF = NF = (K & 0xFF);
					a = (K & 0xFF);
				}
				break;
				
			case 0xF1:
				K = ((page[(wk = (pc++)) >> 13][wk - (wk & msk)] & 0xFF));
				J = (ram[K] & 0xFF);
				J |= (((ram[K + 1] & 0xFF)) << 8);
				J += y;
				I = pce._Rd6502(J);
				if ((p & 0x08) != 0) {
					K = (K & 0xFF00) | (a & 0x0F) - (I & 0x0F) - (~p & 0x01);
					if (((K & 0xFF) & 0x10) != 0)
						K = (K & 0xFF00) | (K & 0xFF) - 6;
					K = (K & 0xFF) | (((a >> 4) - (I >> 4) - ((((K & 0xFF) & 0x10) == 0x10) ? 1 : 0)) << 8);
					if (((K >> 8) & 0x10) != 0)
						K = (K & 0xFF) | (((K >> 8) - 6) << 8);
					a = ((K & 0xFF) & 0x0F) | ((K >> 8) << 4);
					p = (p & ~0x01) | ((((K >> 8) & 0x10) != 0) ? 0 : 0x01);
					ZF = NF = a;
					cycle++;
				} else {
					K = a - I - (~p & 0x01);
					p &= ~0x01;
					p |= (((K >> 8) != 0) ? 0 : 0x01);
					VF = ((a ^ I) & (a ^ (K & 0xFF))) >> 1;
					ZF = NF = (K & 0xFF);
					a = (K & 0xFF);
				}
				break;

			case 0x0A:
				p &= ~0x01;
				p |= a >> 7;
				a = (a << 1) & 0xFF;
				ZF = NF = a;
				break;
				
			case 0x2A:
				K = (K & 0xFF00) | ((a << 1) | (p & 0x01)) & 0xFF;
				p &= ~0x01;
				p |= a >> 7;
				a = (K & 0xFF);
				ZF = NF = a;
				break;
				
			case 0x4A:
				p &= ~0x01;
				p |= a & 0x01;
				a >>= 1;
				ZF = NF = a;
				break;
				
			case 0x6A:
				K = (K & 0xFF00) | (a >> 1) | (p << 7);
				p &= ~0x01;
				p |= a & 0x01;
				a = (K & 0xFF);
				ZF = NF = a;
				break;

			default:
				if (TrapBadOps != 0) {
					debug.TRACE("[M6502 %lX] Unrecognized instruction: $%02X at pc=$%04X\n",
						User, (page[(wk = (pc - 1)) >> 13][wk - (wk & msk)] & 0xFF), (char) (pc - 1));
					Trace = 1;
				}
				break;
			}
			
//			try {
//				out.writeString(
//					"" + (count++)
//					+ " : I=" + I
//					+ " PC=" + pc
//					+ " A=" + a
//					+ " P=" + p
//					+ " X=" + x
//					+ " Y=" + y
//					+ " S=" + s
//					+ " ZF=" + ZF
//					+ " VF=" + VF
//					+ " NF=" + NF
//					+ " IPeriod=" + IPeriod
//					+ " ICount=" + ICount
//					+ " IRequest=" + IRequest
//					+ " IPeriod=" + IPeriod
//					+ " _CycleCount=" + User
//					+ " _AfterCLI=" + AfterCLI
//					+ "\r\n"
//				);
//			} catch (IOException e) {
//			}
			
//			try {
//				out.writeString(pce.io.debugString(count++));
//			} catch (IOException e) {
//			}
			
//			if (User >= 705962)
//			{
//				System.out.println("break!");
//			}

			ICount -= cycle;

			User += cycle;
			
//		} catch (Exception ex) {
//			Debug.printDebugMessage();
//			ex.printStackTrace();
//			return -1;
//		}

			if (ICount <= 0) {
				if (AfterCLI != 0) {
					if ((IRequest & 4) != 0) {
						IRequest &= ~4;
						I = 4;
					} else	if ((IRequest & 1) != 0) {
						IRequest &= ~1;
						I = 1;
					} else if ((IRequest & 8) != 0) {
						IRequest &= ~8;
						I = 8;
					}

					ICount = 0;
					if (IRequest == 0) {
						ICount = IBackup;
						AfterCLI = 0;
					}
				} else {
					I = pce.Loop6502();
					ICount += IPeriod;
				}

//				if (I == 3) {
//					PC = pc;
//					A = a;
//					P = p;
//					X = x;
//					Y = y;
//					S = s;
//					return pc;
//				}
				
				if (I != 0) {
					PC = pc;
					A = a;
					P = p;
					X = x;
					Y = y;
					S = s;
					Int6502(pce, I);
					pc = PC;
					a = A;
					p = P;
					x = X;
					y = Y;
					s = S;
				}

				if ((long) (User - CycleCountOld) > (long) pce.TimerPeriod * 2)
					CycleCountOld = User;
				
				if (pce.isRefreshedScreen) {
					PC = pc;
					A = a;
					P = p;
					X = x;
					Y = y;
					S = s;
					pce.isRefreshedScreen = false;
					return pc;
				}
			} else {
				if (User - CycleCountOld >= pce.TimerPeriod) {
					CycleCountOld += pce.TimerPeriod;
					I = pce.TimerInt();
					if (I != 0) {
						PC = pc;
						A = a;
						P = p;
						X = x;
						Y = y;
						S = s;	
						Int6502(pce, I);
						pc = PC;
						a = A;
						p = P;
						x = X;
						y = Y;
						s = S;
					}
				}
			}

		}
	}
}
