package ixpce;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.util.Vector;

import javax.microedition.io.Connector;

import com.nttdocomo.io.BufferedReader;
import com.nttdocomo.lang.UnsupportedOperationException;
import com.nttdocomo.opt.ui.Graphics2;
import com.nttdocomo.ui.Font;
import com.nttdocomo.ui.IApplication;
import com.nttdocomo.ui.Canvas;
import com.nttdocomo.ui.Graphics;
import com.nttdocomo.ui.Frame;
import com.nttdocomo.ui.Display;


/**
 * @author kenichi
 *
 */
public class Main extends IApplication {

	public void start() {
		MainCanvas mainCanvas = new MainCanvas();
		Display.setCurrent((Frame) (mainCanvas));
		mainCanvas.run();
		mainCanvas.end();
		terminate();
	}
}

/**
 * @author kenichi
 *
 */
class MainCanvas extends Canvas {

	private final M6502 cpu;
	
	private final Pce pce;
	
	private final Sound sound;
	
	private final Debug debug;

	private final Joystick joystick;

	private final static int FPS = 60;

	private final static int INTERVAL = 1000 / FPS;
	
	private long lastTime = 0;
	
	//private PSGi psgi;

	//private NESApu nesApu;
	
	private int messageUpdateCount = 0;
	
	private long appTime = 0;
	
	private int flameCount = 0;
		
	private boolean isMenuCanceled = false;
	
	private int nessageCount = 0;
	
	private String[] message;
	
	private boolean doReset = false;

	private Render render;
	
	private GFont gfont;
	
	private String romFileName = null;
	
	private final String OnStr = "  ON";
	
	private static final int ROM_LIST_PAGE_MAX = 10;
	
	private static final int ROM_LIST_ONE_PAGE_N = 10;

	private String[] romTitles = new String[ROM_LIST_PAGE_MAX * ROM_LIST_ONE_PAGE_N];
	
	private long[] romLastModified = new long[ROM_LIST_PAGE_MAX * ROM_LIST_ONE_PAGE_N];
	
	private int menuX = 48;
	
	private int menuY = 64;
	
	private int romListPage = 0;
	
	/**
	 * RXgN^łB
	 */
	MainCanvas() {
		
		Config.setupDefault();
//		if (this.getWidth() > this.getHeight()) {
//			Config.setupKeysHorizontial();
//		}
				
		this.render = new Render(Pce.WIDTH, Pce.HEIGHT);
		
		// Config
		Config.load();
		
		// ROMLoader
		
		// PAD
		this.joystick = new Joystick(this);
		
		// Debug
		this.debug = new Debug();
		
		// APU
		
		// PPU
		this.sound = new Sound();
		
		// CPU
		this.cpu = new M6502();
		this.pce = new Pce(cpu, sound, joystick, debug, render);
		
		// GFont
		this.gfont = new GFont();

		//nesApu.setNES(nes);
		//setSound();
		
		setBackground(Graphics.getColorOfName(Graphics.BLACK));

		// x\ݒ
		labelVisibleControl();
		
		//cpu.InitSDLog();
	}

	public void end() {
		//cpu.CloseLog();
	}
	
	/* ( Javadoc)
	 * @see com.nttdocomo.ui.Canvas#paint(com.nttdocomo.ui.Graphics)
	 */
	public void paint(Graphics g) {
	}

	/* ( Javadoc)
	 * @see com.nttdocomo.ui.Canvas#processEvent(int, int)
	 */
	public void processEvent(int type, int param) {
	}

	/**
	 * C[vłB
	 */
	public void run() {

		Graphics g = getGraphics();

		// _~[t@C݂ȂꍇA_~[t@C𐶐
		if (!isCreatedDummyFile()) {
			drawMessages(g, new String[] {"CREATING\nDUMMY\nFILES TO\nSD CARD..."}, menuX, menuY);
			createDummyFiles();
		}

		// ROM^Cg擾
		drawMessages(g, new String[] {"GETTING\nROM TITLES..."}, menuX, menuY);
		loadRomTitle();
		
		// ROM̍XV`FbN
		drawMessages(g, new String[] {"CHECK\nROM UPDATED..."}, menuX, menuY);
		checkRomModified();
				
		// ROMXg𐶐
		drawMessages(g, new String[] {"INITIALIZING\nROMS LIST..."}, menuX, menuY);
		createRomTitleList();
		
		// ROM^CgAŏIXVۑ
		saveRomTitle();
		
		// ROMt@CI܂ŁAj[\
		while (romFileName == null) {
			dispMenu(g);
		}
				
		lastTime = System.currentTimeMillis();
		
		int syncWait = 1;
		Graphics2 g2 = null;
		if (g instanceof Graphics2) {
			g2 = (Graphics2) g;
		}
		
		// C[v
		while (true) {
			
			long startAppTime = System.currentTimeMillis();
			
			if (doReset) {
				pce.ResetPCE(cpu);
				doReset = false;
			}
			
			// L[Ԃ擾
			int keyState = getKeypadState();
			
					
			// PADȊÕL[͂ꂽꍇ̏
			if ((keyState & Config.keyMenu) != 0 && !isMenuCanceled) {
				dispMenu(g);
			} else if ((keyState & Config.keyPadSetting) != 0 && !isMenuCanceled) {
				dispPadSetting(g);
			} else if ((keyState & Config.keyStateSave) != 0 && nessageCount == 0) {								
				//stateSave();
			} else if ((keyState & Config.keyStateLoad) != 0 && nessageCount == 0) {
				//stateLoad();
 			} else if ((keyState & Config.keyMenu) == 0) {
				isMenuCanceled = false;
			}
		
			// Emulation-loop
			if (cpu.Run6502(pce) == -1) {
				return;
			}
			
			// OtBbNX`
			g.lock();
			g.setPixels(0, 0, Pce.WIDTH, Pce.HEIGHT, render.getBuffer32(), 0);
			drawMessage(g);
					
			// APȔ
			//nesApu.apu_emulate_frame();

			if (Config.sleepEnable) {
				g.unlock(true);
				sleep();
			} else if (Config.vSyncWaitEnable) {
				if (syncWait <= 1) {
					syncWait = g2.syncUnlock(1);
				} else {
					g.unlock(true);
					sleep();
					syncWait--;
				}
			} else {
				g.unlock(true);
			}
			
			appTime += (System.currentTimeMillis() - startAppTime);
			
			flameCount++;
		}
	}
	
	private boolean isCreatedDummyFile() {
		SDbinding sd = new SDbinding();
		boolean isCreated = false;
		InputStream is = null;
		
		try {
			is = sd.openInputStream("001.PCE");
			if (is != null) {
				isCreated = true;
			}
		} catch (Exception ex) {
		} finally {
			if (is != null) {
				try {
					is.close();
				} catch (IOException ioex) {
				}
			}
			sd.close();
		}
		
		return isCreated;
	}
	
	private void createDummyFiles() {
		SDbinding sd = new SDbinding();
						
        for (int i = 0; i < (ROM_LIST_ONE_PAGE_N * ROM_LIST_PAGE_MAX); i++) {
            sd.createFile(Util.formatNumber(i + 1, 3) + ".PCE");
        }
        
        sd.close();
	}

	private void checkRomModified() {
		SDbinding sd = new SDbinding();

		for (int i = 0; i < ROM_LIST_ONE_PAGE_N * ROM_LIST_PAGE_MAX; i++) {
			// ROMt@C
			String fileName = Util.formatNumber(i + 1, 3) + ".PCE";
			
			// ŏIXV擾
			long last = sd.getLastModified(fileName);
			
			// ŏIXVωꍇA^CgnullZbgāAōĎ擾
			if (last != romLastModified[i]) {
				romTitles[i] = null;
			}
			
			romLastModified[i] = last;
		}
		
		sd.close();
	}
	
	private void saveRomTitle() {
		SDbinding sd = new SDbinding();
		DataOutputStream os = null;
		OutputStreamWriter sw = null;
		
		try {
			os = sd.openDataOutputStream("romtitle.txt");
			sw = new OutputStreamWriter(os);
	
			for (int i = 0; i < ROM_LIST_ONE_PAGE_N * ROM_LIST_PAGE_MAX; i++) {
				sw.write(romTitles[i] + "\t" + romLastModified[i] + "\r\n");		
			}
		} catch (IOException ex) {
			ex.printStackTrace();
		} finally {
			try {
				if (sw != null) {
					sw.close();
				}
				if (os != null) {
					os.close();
				}
			} catch (IOException e) {
			}
		}
		
		sd.close();
	}
	
	private void loadRomTitle() {
		SDbinding sd = new SDbinding();
		DataInputStream is = null;
		InputStreamReader sr = null;
		BufferedReader br = null;
		
		try {
			is = sd.openDataInputStream("romtitle.txt");
			if (is == null) {
				return;
			}
			sr =  new InputStreamReader(is);
			br = new BufferedReader(sr);
			
			for (int i = 0; i < ROM_LIST_ONE_PAGE_N * ROM_LIST_PAGE_MAX; i++) {
				String str = br.readLine();
				if (str == null) {
					break;
				}
				int index = str.indexOf("\t");
				if (index == -1) {
					continue;
				}
				
				romTitles[i] = str.substring(0, index);
				romLastModified[i] = Long.parseLong(str.substring(index + 1));
			}
		} catch (IOException ex) {
			ex.printStackTrace();
		} finally {
			if (is != null) {
				try {
					is.close();
				} catch (IOException e) {
				}
			}
		}
		
		sd.close();
	}
	
	private void createRomTitleList() {
		Resource res = new Resource();
		Vector romList = res.loadRomList();
		SDbinding sd = new SDbinding();

		for (int i = 0; i < ROM_LIST_ONE_PAGE_N * ROM_LIST_PAGE_MAX; i++) {
			
			// ^Cg擾ς݂̏ꍇAXLbv
			if (romTitles[i] != null) {
				continue;
			}
			
			// ROMt@C
			String fileName = Util.formatNumber(i + 1, 3) + ".PCE";
			
			// MD5nbVlvZ
			String hash = Util.getMD5(sd, fileName);
			
			//System.out.println(hash);
			
			// t@C̃nbV̏ꍇXLbv
			if ("348A9791DC41B89796EC3808B5B5262F".equals(hash)) {
				romTitles[i] = "[EMPTY]";
				continue;
			}
			
			// oCiT[`̏
			int lo = 0;
			int hi = romList.size() - 1;
			int mid = (lo + hi) / 2;
			
			// oCiT[`s
			String wk;
			while (lo <= hi && !(wk = ((RomTitle)romList.elementAt(mid)).getHash()).equals(hash)) {
			    if (wk.compareTo(hash) > 0){
			        hi = mid - 1;
			    } else {
			        lo = mid + 1;
			    }
			    mid = (lo + hi) >> 1;
			}
			
			// T[`ʂZbg
			if (lo <= hi) {
				romTitles[i] = ((RomTitle)romList.elementAt(mid)).getRomTitle();
			} else {
				romTitles[i] = "NO NAME";
			}

			sd.close();
		}
	}
	
	private int rapidCount = 0;
	
	public int PadRapidAndHold(int keyState) {
		if (Config.pad1AHold) {
			keyState |= Config.pad1A;
		}
		if (Config.pad1BHold) {
			keyState |= Config.pad1B;
		}
		if (Config.pad1ARapid && (keyState & Config.pad1A) != 0) {
			if (rapidCount != 0) {
				keyState ^= Config.pad1A;
			}
		}
		if (Config.pad1BRapid && (keyState & Config.pad1B) != 0) {
			if (rapidCount != 0) {
				keyState ^= Config.pad1B;
			}
		}
		
		rapidCount = (rapidCount + 1) & 3;
		
		return keyState;
	}
 
	/**
	 * 1/60̎ԊԊȗ҂āAX[v܂B
	 */
	private void sleep() {
		long wait = INTERVAL - (System.currentTimeMillis() - lastTime);
		while (wait > 1) {
			try {
				Thread.sleep(wait / 2);
			} catch (Exception e) {
			}
			wait = INTERVAL - (System.currentTimeMillis() - lastTime);
		}
		lastTime = System.currentTimeMillis();
	}
	
	/**
	 * @param g
	 * @return
	 */
	private boolean existsSyncUnlock(Graphics g) {
		try {			
			if (g instanceof Graphics2) {
				Graphics2 g2 = (Graphics2)g;
				g2.lock();
				g2.syncUnlock(1);
				return true;
			}
		} catch (UnsupportedOperationException e) {
			e.printStackTrace();
		}
		return false;
	}
	
	private String fpsString = "FPS:";
	
	/**
	 * ꎞIɕ\郁bZ[Wʂ֕`悵܂B
	 * @param g
	 */
	private void drawMessage(Graphics g)
	{
		if (Config.dispFPSEnable) {
			if (messageUpdateCount > 60) {
				fpsString = "FPS:" + ((Config.frameSkip + 1) * 1000 * messageUpdateCount / appTime);
				messageUpdateCount = 0;
				appTime = 0;
			}
			messageUpdateCount++;
			
			gfont.DrawString(g, fpsString, 0, 0);
		}
		
		if (nessageCount > 0) {
			drawMessages(g, message, 8, 8);
			nessageCount--;
		}
	}
		
	/**
	 * wʒuɕ`悵܂B
	 * @param g
	 * @param str
	 * @param x
	 * @param y
	 */
	public static void drawString2(Graphics g, String str, int x, int y) {		
		g.clearRect(x, y, 16 * str.length(), 16);

		// 16dotTCỸtHgw
		Font font = Font.getFont(Font.FACE_SYSTEM | Font.STYLE_PLAIN | Font.SIZE_SMALL);
		g.setFont(font);
		g.setColor(Graphics.getColorOfName(Graphics.WHITE));
		g.drawString(str, x, y + font.getAscent());
	}
	
	/**
	 * TEhݒs܂B
	 */
//	private void setSound() {
//		psgi.setFrameBufCount(Config.APUBuffer);
//		psgi.setMasterVolume(Config.APUVolume);
//		psgi.setSoundEnable(Config.APUEnable);
//		psgi.setChEnable(0, true);
//		psgi.setChEnable(1, true);
//		psgi.setChEnable(2, true);
//		psgi.setChEnable(3, true);
//
//		switch (Config.APUMode2) {
//		case 0:
//			psgi.setMode(PSGi.MODE_PRESET);
//			psgi.setToneBank(0, 80);
//			psgi.setToneBank(1, 80);
//			psgi.setToneBank(2, 80);
//			psgi.setToneBank(3, 80);
//			psgi.setToneBank(4, 127);
//			break;
//		case 1:
//			psgi.setMode(PSGi.MODE_PRESET);
//			psgi.setToneBank(0, 81);
//			psgi.setToneBank(1, 81);
//			psgi.setToneBank(2, 81);
//			psgi.setToneBank(3, 81);
//			psgi.setToneBank(4, 127);
//			break;
//		case 2:
//			psgi.setMode(PSGi.MODE_YAMAHA);
//			psgi.setToneBank(0, 0);
//			psgi.setToneBank(1, 1);
//			psgi.setToneBank(2, 2);
//			psgi.setToneBank(3, 3);
//			psgi.setToneBank(4, 4);
//			break;
//		case 3:
//			psgi.setMode(PSGi.MODE_ROHM);
//			psgi.setToneBank(0, 0);
//			psgi.setToneBank(1, 1);
//			psgi.setToneBank(2, 2);
//			psgi.setToneBank(3, 3);
//			psgi.setToneBank(4, 4);
//			break;
//		}
//	}
	
	/**
	 * xi\tgL[j̕\E\ݒs܂B
	 */
	private void labelVisibleControl() {
		try {
			boolean isVisible = Config.dispFPSEnable;
			
			this.setSoftLabelVisible(isVisible);
		} catch (Exception e) { }
	}
	
	/**
	 * Xe[g[hs܂B
	 * @return
	 */
	public boolean stateLoad() {
		DataInputStream input = null;
		try {
			input = Connector.openDataInputStream("scratchpad:///1;pos=0");
			if (!(input.readChar() == 'S' && input.readChar() == 'S')) {
				return false;
			}
			lastTime = input.readLong();
			flameCount = input.readInt();
//			nes.stateLoad(input);
//			ppu.stateLoad(input);
//			pad.stateLoad(input);
		} catch (IOException e) {
			e.printStackTrace();
			return false;
		} finally {
			try {
				input.close();
			} catch (IOException e) {
				e.printStackTrace();
				return false;
			}
		}

		//ppu.isDirtyAll = true;
		message = new String[] {
				"STATE LOADED", 
			};
			nessageCount = 60;
		return true;
	}
		
	/**
	 * Xe[gZ[us܂B
	 * @return
	 */
	public boolean stateSave() {
		DataOutputStream output = null;
		try {
			output = Connector.openDataOutputStream("scratchpad:///1;pos=0");
			output.writeChars("SS");
			output.writeLong(lastTime);
			output.writeInt(flameCount);
//			nes.stateSave(output);
//			ppu.stateSave(output);
//			pad.stateSave(output);
		} catch (IOException e) {
			e.printStackTrace();
			return false;
		} finally {
			try {
				output.close();
			} catch (IOException e) {
				e.printStackTrace();
				return false;
			}
		}
		
		message = new String[] {
				"STATE SAVED", 
			};
		nessageCount = 60;
		return true;
	}
	
	/**
	 * Cj[_CAO\܂B
	 * @param g
	 */
	private void dispMenu(Graphics g)
	{
		isMenuCanceled = false;
		String[] messages = new String[] {
			"@ROM LOAD", 
			"ASCREEN SETTING", 
			"BKEY SETTING", 
			"CSOFT RESET", 
			"DETC", 
			"ERETURN", 
		};
		
		int cursorNo = 1;
		try		
		{
			while (!isMenuCanceled) {
				String[] dispMessage = replaceCursor(messages, cursorNo);
				drawMessages(g, dispMessage, menuX, menuY);
				
				int keyState = getKey();
				if ((keyState & Config.keyOK) != 0) {
					switch (cursorNo) {
						case 1:
							dispRomLoad(g);
							break;
						case 2:
							dispScreenSetting(g);
							break;
//						case 2:
//							dispSoundSetting(g);
//							break;
						case 3:
							dispKeySetting(g);
							break;
						case 4:
							if (confirm2(g)) {
								doReset = true;
							}
							return;
						case 5:
							dispOther(g);
							break;
						case 6: 
							return;
					}
				} else if ((keyState & Config.keyCancel) != 0) {
					return;
				} else if ((keyState & Config.keyMenu) != 0) {
					isMenuCanceled = true;
				} else {
					cursorNo = moveCursor(keyState, cursorNo, 6);
				}
			}
		} finally {
			Config.save();
		}
	}
	
	/**
	 * ROM[h̃_CAO\܂B
	 * @param g
	 */
	private void dispRomLoad(Graphics g)
	{
		int cursorNo = 1;
		
		while (!isMenuCanceled) {
			int offset = romListPage * ROM_LIST_ONE_PAGE_N;
			String[] messages = new String[] {
				"@" + (offset + 1) + ":" + romTitles[offset + 0], 
				"A" + (offset + 2) + ":" + romTitles[offset + 1], 
				"B" + (offset + 3) + ":" + romTitles[offset + 2], 
				"C" + (offset + 4) + ":" + romTitles[offset + 3], 
				"D" + (offset + 5) + ":" + romTitles[offset + 4], 
				"E" + (offset + 6) + ":" + romTitles[offset + 5], 
				"F" + (offset + 7) + ":" + romTitles[offset + 6], 
				"G" + (offset + 8) + ":" + romTitles[offset + 7], 
				"H" + (offset + 9) + ":" + romTitles[offset + 8], 
				"I" + (offset + 10) + ":" + romTitles[offset + 9], 
				"JPREV      KNEXT",
				"    LRETURN", 
			};
		
			String[] dispMessage = replaceCursor(messages, cursorNo);
			drawMessages(g, dispMessage, menuX - 16, menuY);
			
			int keyState = getKey();
			if ((keyState & Config.keyOK) != 0) {
				switch (cursorNo) {
					case 1:
					case 2:
					case 3:
					case 4:
					case 5:
					case 6:
					case 7:
					case 8:
					case 9:
					case 10:
						boolean isError = false;
						try {
							romFileName = Util.formatNumber(offset + cursorNo, 3) + ".PCE";
							if (!pce.InitPCE(romFileName, "")) {
								isError = true;
							}
							g.clearRect(0, 0, getWidth(), getHeight());
							pce.ResetPCE(cpu);
						} catch (Exception ex) {
							isError = true;
						}
						if (isError) {
							alert(g, new String[] {"LOAD ERROR!"});
							break;
						}
						isMenuCanceled = true;
						return;
					case 11:
						if (romListPage > 0) {
							romListPage--;
						}
						break;
					case 12:
						if (romListPage < ROM_LIST_PAGE_MAX - 1) {
							romListPage++;
						}
						break;
					case 13: 
						return;
				}
			} else if ((keyState & Config.keyCancel) != 0) {
				return;
			} else if ((keyState & Config.keyMenu) != 0) {
				isMenuCanceled = true;
			} else {
				cursorNo = moveCursor(keyState, cursorNo, 13);
			}
		}
	}
		
	/**
	 * pbhݒ_CAO\܂B
	 * @param g
	 */
	private void dispPadSetting(Graphics g)
	{
		g.clearClip();
				
		int cursorNo = 1;
		try		
		{
			while (!isMenuCanceled) {
				String[] messages = new String[] {
						"@A HOLD @@" + (Config.pad1AHold ? OnStr : ""), 
						"AB HOLD     " + (Config.pad1BHold ? OnStr : ""), 
						"BA RAPID    " + (Config.pad1ARapid ? OnStr : ""), 
						"CB RAPID    " + (Config.pad1BRapid ? OnStr: ""), 
						"DRETURN", 
					};				
				String[] dispMessage = replaceCursor(messages, cursorNo);
				drawMessages(g, dispMessage, menuX, menuY);
				
				int keyState = getKey();
				if ((keyState & Config.keyOK) != 0) {
					switch (cursorNo) {
						case 1:
							Config.pad1AHold = !Config.pad1AHold;
							break;
						case 2:
							Config.pad1BHold = !Config.pad1BHold;
							break;
						case 3:
							Config.pad1ARapid = !Config.pad1ARapid;
							break;
						case 4:
							Config.pad1BRapid = !Config.pad1BRapid;
							break;
						case 5: 
							return;
					}
				} else if ((keyState & Config.keyCancel) != 0) {
					return;
				} else if ((keyState & Config.keyPadSetting) != 0) {
					isMenuCanceled = true;
				} else {
					cursorNo = moveCursor(keyState, cursorNo, 5);
				}
			}
		} finally {
			g.clearRect(0, 0, getWidth(), getHeight());
			Config.save();
		}
	}

	private void dispOther(Graphics g)
	{
		int cursorNo = 1;
		String[] messages = new String[] {
			"@RESET SETTING", 
			"ARETURN", 
		};
		
		while (!isMenuCanceled) {					
			String[] dispMessage = replaceCursor(messages, cursorNo);
			drawMessages(g, dispMessage, menuX, menuY);	
			int keyState = getKey();
			if ((keyState & Config.keyOK) != 0) {
				switch (cursorNo) {
					case 1:
						if (confirm2(g)) {
							Config.setupDefault();
							if (this.getWidth() > this.getHeight()) {
								Config.setupKeysHorizontial();
							}
							Config.save();
						}
						break;
					case 2:
						return;
				}
			} else if ((keyState & Config.keyCancel) != 0) {
				return;
			} else if ((keyState & Config.keyMenu) != 0) {
				isMenuCanceled = true;
			} else {
				cursorNo = moveCursor(keyState, cursorNo, 2);
			}
		}		
	}
	
	/**
	 * ͂ÅmF_CAO\܂B
	 * @param g
	 * @param msg
	 * @return
	 */
	private boolean confirm2(Graphics g, String msg[]) {
		for (int i = 0; i < msg.length; i++) {			
			int len = 12 - msg[i].length();
			for (int j = 0; j < len; j++) {
				msg[i] += "@";
			}
		}
		
		int cursorNo = 2;
		while (true) {
			String[] messages = new String[4 + msg.length];
			int cnt = 0;
			for (int i = 0; i < msg.length; i++) {
				messages[cnt++] = msg[i];
			}
			messages[cnt++] = "@OK" ;
			messages[cnt++] = "ACANCEL" ;
			
			String[] dispMessage = replaceCursor(messages, cursorNo);
			drawMessages(g, dispMessage, menuX, menuY);	
			int keyState = getKey();
			if ((keyState & Config.keyOK) != 0) {
				switch (cursorNo) {
					case 1:
						return true;
					case 2:
						return false;
				}
			} else if ((keyState & Config.keyCancel) != 0) {
				return false;
			} else if ((keyState & Config.keyMenu) != 0) {
				return false;
			} else {
				cursorNo = moveCursor(keyState, cursorNo, 2);
			}
		}
	}
	
	/**
	 * mF_CAO\܂B
	 * @param g
	 * @param msg
	 */
	private void alert(Graphics g, String msg[]) {
		for (int i = 0; i < msg.length; i++) {
			int len = 12 - msg[i].length();
			for (int j = 0; j < len ; j++) {
				msg[i] += "@";
			}
		}
		
		String[] messages = new String[msg.length];
		int cnt = 0;
		for (int i = 0; i < msg.length; i++) {
			messages[cnt++] = msg[i];
		}
		
		drawMessages(g, messages, menuX, menuY);	
		getKey();
	}

	/**
	 * ͂ÅmF_CAO\܂B
	 * @param g
	 * @return
	 */
	private boolean confirm2(Graphics g) {
		int cursorNo = 2;
		while (true) {
			String[] messages = new String[] {
				"ARE YOU SUREH", 
				"@YES", 
				"ANO", 
			};
					
			String[] dispMessage = replaceCursor(messages, cursorNo);
			drawMessages(g, dispMessage, menuX, menuY);	
			int keyState = getKey();
			if ((keyState & Config.keyOK) != 0) {
				switch (cursorNo) {
					case 1:
						return true;
					case 2:
						return false;
				}
			} else if ((keyState & Config.keyCancel) != 0) {
				return false;
			} else if ((keyState & Config.keyMenu) != 0) {
				return false;
			} else {
				cursorNo = moveCursor(keyState, cursorNo, 2);
			}
		}
	}

	/**
	 * ʐݒ̃_CAO\܂B
	 * @param g
	 */
	private void dispScreenSetting(Graphics g)
	{
		int cursorNo = 1;
		
		while (!isMenuCanceled) {
			String[] messages = new String[] {
				"@FRAME SKIP", 
				"AWAIT", 
				"BDISP FPS" + (Config.dispFPSEnable ? OnStr : "") , 
				"CRETURN", 
			};
		
			String[] dispMessage = replaceCursor(messages, cursorNo);
			drawMessages(g, dispMessage, menuX, menuY);
			
			int keyState = getKey();
			if ((keyState & Config.keyOK) != 0) {
				switch (cursorNo) {
					case 1:
						dispFrameSkipSetting(g);
						break;
					case 2:
						dispSleepSetting(g);
						break;
					case 3:
						Config.dispFPSEnable = !Config.dispFPSEnable;
						if (!Config.dispFPSEnable) {
							setSoftLabel(SOFT_KEY_2, null);
						}
						labelVisibleControl();
						break;
					case 4: 
						return;
				}
			} else if ((keyState & Config.keyCancel) != 0) {
				return;
			} else if ((keyState & Config.keyMenu) != 0) {
				isMenuCanceled = true;
			} else {
				cursorNo = moveCursor(keyState, cursorNo, 4);
			}
		}
	}
	
	/**
	 * t[XLbvݒ_CAO\܂B
	 * @param g
	 */
	private void dispFrameSkipSetting(Graphics g)
	{
		int cursorNo = 1;
		
		while (!isMenuCanceled) {
			String[] messages = new String[] {
				"@FS1" + getFrameSkipStr(0), 
				"AFS2" + getFrameSkipStr(1), 
				"BFS3" + getFrameSkipStr(2), 
				"CFS4" + getFrameSkipStr(3), 
				"DFS5" + getFrameSkipStr(4), 
				"EFS6" + getFrameSkipStr(5), 
				"FFS7" + getFrameSkipStr(6), 
				"GFS8" + getFrameSkipStr(7), 
				"HRETURN", 
			};
		
			String[] dispMessage = replaceCursor(messages, cursorNo);
			drawMessages(g, dispMessage, menuX, menuY);
			
			int keyState = getKey();
			if ((keyState & Config.keyOK) != 0) {
				switch (cursorNo) {
					case 1:
					case 2:
					case 3:
					case 4:
					case 5:
					case 6:
					case 7:
					case 8:
						Config.frameSkip = cursorNo - 1;
						break;
					case 9: 
						return;
				}
			} else if ((keyState & Config.keyCancel) != 0) {
				return;
			} else if ((keyState & Config.keyMenu) != 0) {
				isMenuCanceled = true;
			} else {
				cursorNo = moveCursor(keyState, cursorNo, 9);
			}
		}
	}
	
	/**
	 * ݒlƈvꍇON̕Ԃ܂B
	 * @param skip
	 * @return
	 */
	private String getFrameSkipStr(int skip) {
		return (Config.frameSkip == skip) ? OnStr : "";
	}
	
	/**
	 * X[vݒ_CAO\܂B
	 * @param g
	 */
	private void dispSleepSetting(Graphics g)
	{
		int cursorNo = 1;
		
		while (!isMenuCanceled) {
			String[] messages = new String[] {
				"@SLEEP WAIT    " + (Config.sleepEnable ? OnStr : ""), 
				"AV-SYNC WAIT   " + (Config.vSyncWaitEnable ? OnStr : ""), 
				"BRETURN", 
			};
		
			String[] dispMessage = replaceCursor(messages, cursorNo);
			drawMessages(g, dispMessage, menuX, menuY);
			
			int keyState = getKey();
			if ((keyState & Config.keyOK) != 0) {
				switch (cursorNo) {
					case 1:
						Config.sleepEnable = !Config.sleepEnable;
						if (Config.sleepEnable) {
							Config.vSyncWaitEnable = false;
						}
						break;
					case 2: 
						Config.vSyncWaitEnable = !Config.vSyncWaitEnable;
						if (Config.vSyncWaitEnable && !existsSyncUnlock(g)) {
							Config.vSyncWaitEnable = false;
							dispError(g);
						}
						if (Config.vSyncWaitEnable) {
							Config.sleepEnable = false;
						}
						break;
					case 3: 
						return;
				}
			} else if ((keyState & Config.keyCancel) != 0) {
				return;
			} else if ((keyState & Config.keyMenu) != 0) {
				isMenuCanceled = true;
			} else {
				cursorNo = moveCursor(keyState, cursorNo, 3);
			}
		}
	}
	
	/**
	 * Ή@\̃_CAO\܂B
	 * @param g
	 */
	private void dispError(Graphics g) {
		String[] messages = new String[] {
			"NOT IMPLEMENTED METHOD", 
		};
				
		drawMessages(g, messages, menuX, menuY);			
		getKey();
	}
	
//	/**
//	 * TEhݒ_CAO\܂B
//	 * @param g
//	 */
//	private void dispSoundSetting(Graphics g)
//	{
//		int cursorNo = 1;
//
//		while (!isMenuCanceled) {
//			String[] messages = new String[] {
//				"",
//				"@񂯁J񃂁[gJ@", 
//				"AzJ[@@@@", 
//				"BnJbt@@@@@@", 
//				"CƂ䂤@@@" + (Config.APUEnable ? "d" : "@") + "", 
//				"DXeI@@@@@" + (Config.APUStereoEnable ? "d" : "@") + "", 
//				"EƁJ@@@@@@", 
//				"", 		
//			};
//		
//			String[] dispMessage = replaceCursor(messages, cursorNo);
//			drawMessages(g, dispMessage, menuX, menuY);
//			
//			int keyState = getKey();
//			if ((keyState & Config.keyOK) != 0) {
//				switch (cursorNo) {
//					case 1:	
//						dispSoundEngineSetting(g);
//						break;
//					case 2: 
//						dispSoundVolumeSetting(g);
//						break;
//					case 3:
//						dispSoundBufferSetting(g);
//						break;
//					case 4: 
//						Config.APUEnable = !Config.APUEnable;
//						setSound();
//						break;
//					case 5: 
//						Config.APUStereoEnable = !Config.APUStereoEnable;
//						setSound();
//						break;
//					case 6:
//						return;
//				}
//			} else if ((keyState & Config.keyCancel) != 0) {
//				return;
//			} else if ((keyState & Config.keyMenu) != 0) {
//				isMenuCanceled = true;
//			} else {
//				cursorNo = moveCursor(keyState, cursorNo, 6);
//			}
//		}
//	}
//	
//	/**
//	 * TEh[hݒ_CAO\܂B
//	 * @param g
//	 */
//	private void dispSoundEngineSetting(Graphics g)
//	{
//		int cursorNo = 1;
//		
//		while (!isMenuCanceled) {
//			String[] messages = new String[] {
//				"",
//				"@[gJP@@@@" + getSoundModeStr(0) + "", 
//				"A[gJQ@@@@" + getSoundModeStr(1) + "", 
//				"B[gJR@@@@" + getSoundModeStr(2) + "", 
//				"C[gJS@@@@" + getSoundModeStr(3) + "", 
//				"DƁJ@@@@@@", 
//				"", 		
//			};
//		
//			String[] dispMessage = replaceCursor(messages, cursorNo);
//			drawMessages(g, dispMessage, menuX, menuY);
//			
//			int keyState = getKey();
//			if ((keyState & Config.keyOK) != 0) {
//				switch (cursorNo) {
//					case 1:	
//					case 2: 
//					case 3: 
//					case 4: 
//						Config.APUMode2 = cursorNo - 1;
//						//setSound();
//						break;
//					case 5:
//						return;
//				}
//			} else if ((keyState & Config.keyCancel) != 0) {
//				return;
//			} else if ((keyState & Config.keyMenu) != 0) {
//				isMenuCanceled = true;
//			} else {
//				cursorNo = moveCursor(keyState, cursorNo, 5);
//			}
//		}
//	}
//	
//	/**
//	 * TEh[h̐ݒlƈvꍇɁA"d"̕Ԃ܂B
//	 * @param soundMode
//	 * @return
//	 */
//	private String getSoundModeStr(int soundMode) {
//		return (Config.APUMode2 == soundMode) ? "d" : "@";
//	}
//	
//	/**
//	 * ʐݒ_CAO\܂B
//	 * @param g
//	 */
//	private void dispSoundVolumeSetting(Graphics g)
//	{
//		int cursorNo = 1;
//		
//		while (!isMenuCanceled) {
//			String[] messages = new String[] {
//				"",
//				"@zJ[QO@" + getVolumeStr(20 ) + "", 
//				"AzJ[SO@" + getVolumeStr(40 ) + "", 
//				"BzJ[UO@" + getVolumeStr(60 ) + "", 
//				"CzJ[WO@" + getVolumeStr(80 ) + "", 
//				"DzJ[l`w@" + getVolumeStr(100) + "", 
//				"EƁJ@@@@@@@", 
//				"", 		
//			};
//		
//			String[] dispMessage = replaceCursor(messages, cursorNo);
//			drawMessages(g, dispMessage, menuX, menuY);
//			
//			int keyState = getKey();
//			if ((keyState & Config.keyOK) != 0) {
//				switch (cursorNo) {
//					case 1:	
//					case 2: 
//					case 3: 
//					case 4: 
//					case 5: 
//						Config.APUVolume = cursorNo * 20;
//						setSound();
//						break;
//					case 6:
//						return;
//				}
//			} else if ((keyState & Config.keyCancel) != 0) {
//				return;
//			} else if ((keyState & Config.keyMenu) != 0) {
//				isMenuCanceled = true;
//			} else {
//				cursorNo = moveCursor(keyState, cursorNo, 6);
//			}
//		}
//	}
//	
//	/**
//	 * ʐݒ肪ݒlƈvꍇɁA"d"̕Ԃ܂B
//	 * @param vol
//	 * @return
//	 */
//	private String getVolumeStr(int vol) {
//		return (Config.APUVolume == vol) ? "d" : "@";
//	}
//	
//	/**
//	 * TEhobt@̐ݒ_CAO\܂B
//	 * @param g
//	 */
//	private void dispSoundBufferSetting(Graphics g)
//	{
//		int cursorNo = 1;
//
//		while (!isMenuCanceled) {
//			String[] messages = new String[] {
//				"",
//				"@nJbt@T@@@@"+ getApuBufferStr( 5) + "", 
//				"AnJbt@PO@@@"+ getApuBufferStr(10) + "", 
//				"BnJbt@PT@@@"+ getApuBufferStr(15) + "", 
//				"CnJbt@QO@@@"+ getApuBufferStr(20) + "", 
//				"DnJbt@QT@@@"+ getApuBufferStr(25) + "", 
//				"EnJbt@RO@@@"+ getApuBufferStr(30) + "", 
//				"FnJbt@RT@@@"+ getApuBufferStr(35) + "", 
//				"GnJbt@SO@@@"+ getApuBufferStr(40) + "", 
//				"HƁJ@@@@@@@", 
//				"", 		
//			};
//				
//			String[] dispMessage = replaceCursor(messages, cursorNo);
//			drawMessages(g, dispMessage, menuX, menuY);
//			
//			int keyState = getKey();
//			if ((keyState & Config.keyOK) != 0) {
//				switch (cursorNo) {
//					case 1:	
//					case 2: 
//					case 3:
//					case 4:
//					case 5:
//					case 6:
//					case 7:
//					case 8:
//						Config.APUBuffer = cursorNo * 5;
//						setSound();
//						break;
//					case 9:
//						return;
//				}
//			} else if ((keyState & Config.keyCancel) != 0) {
//				return;
//			} else if ((keyState & Config.keyMenu) != 0) {
//				isMenuCanceled = true;
//			} else {
//				cursorNo = moveCursor(keyState, cursorNo, 9);
//			}
//		}
//	}
//	
//	/**
//	 * TEhobt@̐ݒlƈvꍇɁA"d"̕Ԃ܂B
//	 * @param bufSize
//	 * @return
//	 */
//	private String getApuBufferStr(int bufSize) {
//		return (Config.APUBuffer == bufSize) ? "d" : "@";
//	}
	
	/**
	 * L[ݒ̃_CAO\܂B
	 * @param g
	 */
	private void dispKeySetting(Graphics g)
	{
		int cursorNo = 1;

		while (!isMenuCanceled) {
			String[] messages = new String[] {
				"@UP     " + getKeySettingString(Config.pad1Up), 
				"ADOWN   " + getKeySettingString(Config.pad1Down), 
				"BLEFT   " + getKeySettingString(Config.pad1Left), 
				"CRIGHT  " + getKeySettingString(Config.pad1Right), 
				"DA      " + getKeySettingString(Config.pad1A), 
				"EB      " + getKeySettingString(Config.pad1B), 
				"FSTART  " + getKeySettingString(Config.pad1Start), 
				"GSELECT " + getKeySettingString(Config.pad1Select), 
				"HSAVE   " + getKeySettingString(Config.keyStateSave), 
				"ILOAD   " + getKeySettingString(Config.keyStateLoad), 
				"JRETURN", 
			};
			
			String[] dispMessage = replaceCursor(messages, cursorNo);
			drawMessages(g, dispMessage, menuX, menuY);
			
			int keyState = getKey();
			if ((keyState & Config.keyOK) != 0) {
				switch (cursorNo) {
					case 1:	
					case 2: 
					case 3:
					case 4:
					case 5:
					case 6:
					case 7:
					case 8:
					case 9:
					case 10:
						dispKeySetting2(g, cursorNo);
						break;
					case 11:
						return;
				}
			} else if ((keyState & Config.keyCancel) != 0) {
				return;
			} else if ((keyState & Config.keyMenu) != 0) {
				isMenuCanceled = true;
			} else {
				cursorNo = moveCursor(keyState, cursorNo, 11);
			}
		}
	}
			
	/**
	 * L[I_CAO\܂B
	 * @param g
	 * @param keyNo
	 */
	private void dispKeySetting2(Graphics g, int keyNo) {
		String[] messages = new String[] {
			"PUSH A KEY", 
		};
				
		drawMessages(g, messages, menuX, menuY);			
		int keyState = getPadKey();
		switch (keyNo) {
			case 1: Config.pad1Up = keyState; break;
			case 2: Config.pad1Down = keyState; break;
			case 3: Config.pad1Left = keyState; break;
			case 4: Config.pad1Right = keyState; break;
			case 5: Config.pad1A = keyState; break;
			case 6: Config.pad1B = keyState; break;
			case 7: Config.pad1Start = keyState; break;
			case 8: Config.pad1Select = keyState; break;
			case 9: Config.keyStateSave = keyState; break;
			case 10: Config.keyStateLoad = keyState; break;
		}
	}
	
	/**
	 * w肵L[ǂݎlɑΉA\Ԃ܂B
	 * @param padkey
	 * @return
	 */
	private String getKeySettingString(int padkey) {
		switch (padkey) {
			case (1 << Display.KEY_0): 			return "       0";
			case (1 << Display.KEY_1): 			return "       1";
			case (1 << Display.KEY_2): 			return "       2";
			case (1 << Display.KEY_3): 			return "       3";
			case (1 << Display.KEY_4): 			return "       4";
			case (1 << Display.KEY_5): 			return "       5";
			case (1 << Display.KEY_6): 			return "       6";
			case (1 << Display.KEY_7): 			return "       7";
			case (1 << Display.KEY_8): 			return "       8";
			case (1 << Display.KEY_9): 			return "       9";
			case (1 << Display.KEY_ASTERISK): 	return "       *";
			case (1 << Display.KEY_DOWN): 		return "    DOWN";
			case (1 << Display.KEY_IAPP): 		return "     APP";
			case (1 << Display.KEY_LEFT): 		return "    LEFT";
			case (1 << Display.KEY_PAGE_DOWN): 	return "   PDOWN";
			case (1 << Display.KEY_PAGE_UP): 	return "     PUP";
			case (1 << Display.KEY_POUND): 		return "       #";
			case (1 << Display.KEY_RIGHT): 		return "   RIGHT";
			case (1 << Display.KEY_SELECT):		return "  SELECT";
			case (1 << Display.KEY_UP): 		return "      UP";
		}
		return "";
	}
	
	/**
	 * j[ɎgpL[ǂݎ܂B
	 * @return
	 */
	private int getKey() {
		final int mask = 
			  Config.keyUp
			| Config.keyDown
			| Config.keyOK
			| Config.keyCancel
			| Config.keyMenu;

		int keyState;
		
		do {
			keyState = getKeypadState();
			sleep();
		} while ((keyState & mask) != 0);
		
		do {
			keyState = getKeypadState();
			sleep();
		} while (bitCount(keyState & mask) != 1);
		
		return keyState;
	}
	
	/**
	 * pbhݒɎgpL[ǂݎ܂B
	 * @return
	 */
	private int getPadKey() {
		final int mask = 
			(1 << Display.KEY_0)
		  | (1 << Display.KEY_1)
		  | (1 << Display.KEY_2)
		  | (1 << Display.KEY_3)
		  | (1 << Display.KEY_4)
		  | (1 << Display.KEY_5)
		  | (1 << Display.KEY_6)
		  | (1 << Display.KEY_7)
		  | (1 << Display.KEY_8)
		  | (1 << Display.KEY_9)
		  | (1 << Display.KEY_ASTERISK)
		  | (1 << Display.KEY_DOWN)
		  | (1 << Display.KEY_IAPP)
		  | (1 << Display.KEY_LEFT)
		  | (1 << Display.KEY_PAGE_DOWN)
		  | (1 << Display.KEY_PAGE_UP)
		  | (1 << Display.KEY_POUND)
		  | (1 << Display.KEY_RIGHT)
		  | (1 << Display.KEY_SELECT)
		  | (1 << Display.KEY_UP);

		int keyState;
		
		do {
			keyState = getKeypadState();
			sleep();
		} while ((keyState & mask) != 0);
		
		do {
			keyState = getKeypadState();
			sleep();
		} while (bitCount(keyState & mask) != 1);
		
		return keyState;
	}
	
	/**
	 * w肵intϐ̃rbgĂ鐔Ԃ܂B
	 * @param bit
	 * @return
	 */
	private int bitCount(int bit) {
		int count = 0;
		for (int i = 0; i < 32; i++) {
			if ((bit & (1 << i)) != 0) {
				count++;
			}
		}
		return count;
	}
		 
	/**
	 * J[\ړ̏s܂B
	 * @param keyState
	 * @param cursorNo
	 * @param maxNo
	 * @return
	 */
	private int moveCursor(int keyState, int cursorNo, int maxNo) {
		
		if ((keyState & Config.keyUp) != 0)
		{
			cursorNo--;
		}
		if ((keyState & Config.keyDown) != 0)
		{
			cursorNo++;
		}
		
		if (cursorNo > maxNo) {
			cursorNo = 1;
		} else if (cursorNo < 1) {
			cursorNo = maxNo;
		}
		
		return cursorNo;
	}
	
	/**
	 * w肵J[\NOɑΉuJ[\Lɒu܂B
	 * @param messages
	 * @param cursorNo
	 * @return
	 */
	private String[] replaceCursor(String[] messages, int cursorNo) {
		final char[] numChars = new char[] {
			'@','@','A','B','C','D','E','F','G','H','I','J','K','L','M','N',
			'O','P','Q','R','S'
		};
			
		String[] ret = new String[messages.length];
		
		for (int i = 0; i < messages.length; i++) {
			ret[i] = messages[i].replace(numChars[cursorNo], '>');
			for (int j = 1; j < numChars.length; j++) {
				ret[i] = ret[i].replace(numChars[j], '@');
			}
		}		
		return ret;
	}
	
	/**
	 * w肵zʂɕ`悵܂B
	 * @param g
	 * @param messages
	 * @param x
	 * @param y
	 */
	private void drawMessages(Graphics g, String[] messages, int x, int y) {
		g.lock();
		gfont.drawBackGround(g, getWidth(), getHeight());
		for (int i = 0; i < messages.length; i++) {
			//drawString2(g, messages[i], x, y += 16);
			gfont.DrawString(g, messages[i], x, y += gfont.getFontWidth());
		}
		g.unlock(true);
	}
}
